/*
 * The "Memulator" timing-accurate storage emulation environment
 * Authors: John Linwood Griffin, John Bucy, Greg Ganger
 *
 * Copyright (c) of Carnegie Mellon University, 2001-2004.
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 *
 */

#define MEMULATOR_EVENT_TARGET_SIMULATOR_TIME    (0x40)

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <aio.h>
#include <errno.h>
#include <fcntl.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/event.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/time.h>
#include <sys/queue.h>
#include <sched.h>

#include "memulator.h"
#define USE_SIMULATION
#ifdef USE_SIMULATION
#include "disksim/src/disksim_interface.h"
#endif

static int debug=0;
static int kq_fd=-1;  /* Descriptor for kernel event queue */
static int memulator_sock=-1;  /* Socket for IPC with memulator */
static int received_termination=0;  /* When memulator tells us to shut down */

static char *calibration_prefix = NULL;
static int calibration_fd = -1;

static char *skew=0;

/************************************************************************
 * Support functions
 ************************************************************************/

#define log(a,b...) \
   do { fprintf(stdout, "%s: %s: " a, "clock", __func__, b); \
        fflush(stdout); } while (0)


static inline void
time_stamp (u_int64_t *d)
{
  __asm__ __volatile__ ("rdtsc" : "=&A" (*d));
}


void *
xcalloc (size_t size)
{
  register void *value = calloc(1, size);
  if (value == 0) {
    log("unable to allocate %d bytes, exiting", size);
    assert(0);
  }
  return value;
}


/************************************************************************
 * Timings output
 ************************************************************************/

void
open_calibration_times_file (void)
{
  const int filename_maxlen=80;
  char filename[filename_maxlen];

  if (!calibration_prefix) return;

  if ((strlen(calibration_prefix) + 4) > filename_maxlen) {
    log("please increase filename_maxlen (%d)\n", filename_maxlen);
    assert(0);
  }

  snprintf(filename, filename_maxlen,
	   "%s.sim",
	   calibration_prefix);

  calibration_fd = open(filename, O_CREAT|O_WRONLY, 0644);
  if (calibration_fd == -1) {
    log("unable to open %s: %s\n", filename, strerror(errno));
    assert(0);
  }
}


void
close_calibration_times_file (void)
{
  if (calibration_fd == -1) return;

  if (close(calibration_fd) == -1) {
    log("close: %s\n", strerror(errno));
  }
  calibration_fd = -1;
}


void
calibration_time_stamp (double time, struct disksim_request *r)
{
  const int buflen = 160;
  unsigned char buf[buflen];
  int bytes;

  if (calibration_fd == -1) return;

  bytes = snprintf(buf, buflen,
		   "%ld\t%d\t%x\t%x\t%f\t%d\n",
		   r->blkno,
		   r->bytecount / 512,
		   (r->flags & DISKSIM_READ) ? 0 : 1,  /* 0 for read */
		   MEMULATOR_EVENT_TARGET_SIMULATOR_TIME,
		   time - r->start,
		   (int)r->reqctx);

  write(calibration_fd, buf, bytes);
}


/************************************************************************
 * Interprocess communication
 ************************************************************************/

static int
send_ipc (int sock, void *buf, size_t len)
{
  u_int8_t *msg;
  int bytes = 0;	/* Total bytes sent */
  int rc;

  if (debug) log("%s\n","send_ipc");

  msg = buf;

  while (bytes < len) {
    if ((rc = send(sock, msg + bytes, len - bytes, 0)) == -1) {
      log("send: %s", strerror(errno));
      return (-1);
    }
    bytes += rc;
  }

  if (bytes != len) {
    log("bad transfer (%d != %d)\n", bytes, len);
  }

  return (bytes);
}


static int
recv_ipc (int sock, void *buf, size_t len)
{
  u_int8_t *msg;
  int bytes = 0;	/* Total bytes received */
  int rc;

  if (debug) log("%s\n","recv_ipc");

  msg = buf;

  while (bytes < len) {
    if ((rc = recv(sock, buf + bytes, len - bytes, 0)) == -1) {
      log("recv: %s", strerror(errno));
      return (-1);
    }
    bytes += rc;
  }

  if (bytes != len) {
    log("bad transfer (%d != %d)\n", bytes, len);
  }

  return (bytes);
}


static void
memulator_attach (void)
{
  struct sockaddr_un sun;
  size_t size;
  ipc_message_t msg_out;

  assert(memulator_sock==-1);

  /* socket */

  if ((memulator_sock = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) {
    log("socket: %s", strerror(errno));
    assert(0);
  }

  /* connect */

  memset(&sun, 0, sizeof(struct sockaddr_un));
  sun.sun_family = AF_UNIX;
  strncpy(sun.sun_path, MEMULATOR_SOCKET_FILENAME, 104);
  size = SUN_LEN(&sun);
  if (connect(memulator_sock, (struct sockaddr *)&sun, size) == -1) {
    log("connect: %s\n", strerror(errno));
    assert(0);
  }

  /* Send connect message */

  memset(&msg_out, 0, sizeof(ipc_message_t));
  msg_out.magic  = IPC_MAGIC;
  msg_out.source = IPC_SOURCE_CLOCK;
  msg_out.type   = IPC_TYPE_CONNECT;
  send_ipc(memulator_sock, &msg_out, sizeof(ipc_message_t));
}


/************************************************************************
 * Simulation support
 ************************************************************************/

static struct disksim_interface *disksim_interface;
static int num_pending_requests=0;

/* GRONK -- assumes nc94.
 * 
 * nc93: 1993907191
 * nc94: 1993868287
 * nc95: 1993925703
 * nc96: 1993905323
 */
const u_int64_t ts_calibration = 1993868287LL; /* correct */
// const double    ts_calibration_d = 1993868287.0;
u_int64_t ts_simulation_start = 0;

double prev_simtime_msec = 0.0;
double next_simtime_msec = 0.0;
u_int64_t next_callback_ts = 0;
disksim_interface_callback_t callback_fn = NULL;

struct disksim_request *completed_request = NULL;
double completed_time_msec = 0.0;
u_int64_t completed_time_ts = 0;


static inline double
get_curtime_msec (void)
{
  u_int64_t ts_now;
  u_int64_t ts_elapsed;

  time_stamp(&ts_now);
  ts_elapsed = ts_now - ts_simulation_start;

  /* convert to msec with the multiply-by-1000 */
  return ((double)(ts_elapsed) / (double)ts_calibration) * 1000.0;
}


/* callback (disksim -> client) to register a callback (client ->
 * disksim); when simulated time reaches t, client code should invoke
 * fn. */
void
schedule_callback (disksim_interface_callback_t fn,
		   double time_msec,
		   void *context)
{
  //fn(disksim_interface, time_msec, 0);
  //return;

  next_simtime_msec = time_msec;
  callback_fn = fn;
#if 0
  next_callback_ts = 
    ts_simulation_start + ((time_msec - 0.3) / 1000.0) * ts_calibration;
#endif
}


/* cancel the pending callback set for t */
void
deschedule_callback (double time_msec,
		     void *context)
{
  next_simtime_msec = 0.0;
  callback_fn = NULL;
  next_callback_ts = 0;

#if 0
  callback_ts = 0;
  callback_simtime = 0.0;
  callback_fn = NULL;
#endif
}


/* GRONK: fix me! */
int sk_len=0;
int sk_typ=0;

void
handle_request_arrive (ipc_message_t *msg_in)
{
  struct disksim_request *r;
  double interface_start;

  if (debug) log("request %d arrival\n", msg_in->serial);

  r = xcalloc(sizeof(struct disksim_request));
  if (msg_in->type == IPC_TYPE_READ) {
    r->flags |= DISKSIM_READ;
  } else if (msg_in->type == IPC_TYPE_WRITE) {
    r->flags |= DISKSIM_WRITE;
  } else {
    assert(0);
  }
  r->devno          = 0;
  r->blkno          = msg_in->request_offset_blocks;
  r->bytecount      = msg_in->request_length_blocks * 512;
  r->start          = get_curtime_msec();
  r->batchno        = 0;
  r->batch_complete = 0;
  r->reqctx         = (void *)msg_in->serial;

  num_pending_requests++;

  if (prev_simtime_msec > r->start) {
    interface_start = prev_simtime_msec;
  } else {
    interface_start = r->start;
  }

  sk_len = r->bytecount / 512;
  sk_typ = r->flags;

  disksim_interface_request_arrive(disksim_interface, interface_start, r);
}


double compl_min = 99999.9;
double compl_max = -99999.9;
double compl_total = 0.0;
int compl_num = 0;
int compl_num_neg = 0;

/* callback when requests complete */
void
handle_request_complete (double time,
			 struct disksim_request *r,
			 void *context)
{
#if 0
  double diff = time - get_curtime_msec();
  if (diff < compl_min) compl_min = diff;
  if (diff > compl_max) compl_max = diff;
  compl_total += diff;
  compl_num++;
#endif

  completed_request = r;
  completed_time_msec = time;
#if 0
  completed_time_ts =
    ts_simulation_start + ((time - 0.2) / 1000.0) * ts_calibration;
#endif
}


void
send_completion (double time, struct disksim_request *r)
{
  ipc_message_t msg_out;

  if (debug) log("request %d complete\n", (int)r->reqctx);

  memset(&msg_out, 0, sizeof(ipc_message_t));
  msg_out.magic  = IPC_MAGIC;
  msg_out.source = IPC_SOURCE_CLOCK;
  msg_out.type   = IPC_TYPE_REQUEST_OK;
  msg_out.serial = (int)r->reqctx;
  send_ipc(memulator_sock, &msg_out, sizeof(ipc_message_t));

  sched_yield();

  calibration_time_stamp(time, r);

  // printf("error = %f\n", get_simtime_msec() - time);

  num_pending_requests--;
  free(r);  r = NULL;
}



static void
start_simulation (void)
{
  time_stamp(&ts_simulation_start);

  disksim_interface =
    disksim_interface_initialize("parfile",
				 "outfile",
				 handle_request_complete,
				 schedule_callback,
				 deschedule_callback,
				 0);
}


static void
stop_simulation (void)
{
  disksim_interface_shutdown(disksim_interface, get_curtime_msec());
}


/************************************************************************
 * Event processing
 ************************************************************************/

static void
handle_eof (int sock)
{
  const int max_events=1;
  struct kevent events[max_events];
  int n=0;

  assert(sock==memulator_sock);

  EV_SET(&events[n++], sock, EVFILT_READ, EV_DELETE, 0,0,0);
  assert(n<=max_events);
  if (kevent(kq_fd, events, n, NULL, 0, NULL) == -1) {
    log("kevent: %s\n", strerror(errno));
    assert(0);
  }

  if (sock == memulator_sock) {
    if (close(memulator_sock) == -1) {
      log("close: %s\n", strerror(errno));
    }
    memulator_sock = -1;
    received_termination = 1;
  }
  else {
    log("unknown socket %d\n", sock);
  }
}


static void
handle_ipc (int sock)
{
  ipc_message_t msg_in;

  assert(sock==memulator_sock);

  memset(&msg_in, 0, sizeof(ipc_message_t));
  recv_ipc(sock, &msg_in, sizeof(ipc_message_t));

  if (msg_in.magic != IPC_MAGIC) {
    log("bad magic %#x\n", msg_in.magic);
    return;
  }

  if (msg_in.source != IPC_SOURCE_MEMULATOR) {
    log("bad source %#x\n", msg_in.source);
    return;
  }

  switch (msg_in.type) {

  case IPC_TYPE_READ:
  case IPC_TYPE_WRITE:
    handle_request_arrive(&msg_in);
    break;

  case IPC_TYPE_DISCONNECT:
    log("%s\n", "Received disconnection directive.");
    handle_eof(sock);
    break;

  default:
    log("unknown type %d\n", msg_in.type);
  }
}


/************************************************************************
 * Main loop
 ************************************************************************/

static int
main_loop (void)
{
  const int max_events=16;
  struct kevent events[max_events];
  int n=0;

  /* Register kqueue for event notification */
  if ((kq_fd = kqueue()) == -1) {
    log("kqueue: %s\n", strerror(errno));
    assert(0);
  }

  signal(SIGHUP, SIG_IGN);
  signal(SIGINT, SIG_IGN);
  signal(SIGTERM, SIG_IGN);
  signal(SIGPROF, SIG_IGN);
  signal(SIGALRM, SIG_IGN);
  signal(SIGSTOP, SIG_IGN);
  signal(SIGTSTP, SIG_IGN);

  /* Set up some default events */
  //EV_SET(&events[n++], SIGHUP,  EVFILT_SIGNAL, EV_ADD, 0,0,0);
  //EV_SET(&events[n++], SIGINT,  EVFILT_SIGNAL, EV_ADD, 0,0,0);
  //EV_SET(&events[n++], SIGTERM, EVFILT_SIGNAL, EV_ADD, 0,0,0);
  EV_SET(&events[n++], SIGALRM, EVFILT_SIGNAL, EV_ADD, 0,0,0);
  if (memulator_sock != -1) {
    EV_SET(&events[n++], memulator_sock, EVFILT_READ, EV_ADD, 0,0,0);
  }

  assert(n<=max_events);
  if (kevent(kq_fd, events, n, NULL, 0, NULL) == -1) {
    log("kevent 1: %s\n", strerror(errno));
    assert(0);
  }

 while (!received_termination) {
    int num_events, i;
    struct timespec ts = {0, 0};

    /* Check for the next signal or read ready */
    /* GRONK: example code uses clever value for timeout */

    num_events = kevent(kq_fd, NULL, 0, events, max_events, &ts);
    if (num_events == -1) {
      if (errno == EINTR) continue;
      log("kevent 2: %s\n", strerror(errno));
      assert(0);
    }
    if (num_events > max_events) {
      log("kevent 3: %s\n", strerror(errno));
      assert(0);
    }

    /* Process all received events */

    for (i=0; i<num_events; i++) {
      if ((events[i].flags & EV_ERROR) != 0) {
	log("kevent 4: %s\n", strerror(errno));
	assert(0);
      }

      switch (events[i].filter) {

      case EVFILT_READ:
	if (events[i].ident == memulator_sock) {
	  if (!(events[i].flags & EV_EOF)) {
	    handle_ipc(events[i].ident);
	  }
	  else {  /* Other side disconnected */
	    handle_eof(events[i].ident);
	  }
	}
	else {
	  log("EVFILT_READ: unknown ident %#x\n", events[i].ident);
	}
	break;

      case EVFILT_SIGNAL:
	if (debug) log("signal received (%#x)\n", events[i].ident);
	log("Unknown signal received: %#x\n", events[i].ident);
	assert(0);
	break;

      default:
	log("unknown event %#x\n", events[i].filter);
	break;
      }
    }

    {
      double time_msec;
      //u_int64_t time_ts;
      int yield_ok = 1;
      double skew_ms=0.0;

      if (skew) {
	if (*skew == 'A') {
	  if (sk_typ & DISKSIM_READ) {
	    skew_ms = 0.108 + 0.115 + 0.000667 * (sk_len < 8 ? 0 : sk_len - 8);
	  } else {
	    skew_ms = 0.108 + 0.115 + 0.000775 * (sk_len < 8 ? 0 : sk_len - 8);
	  }
	}
	else if (*skew == 'Q') {
	  if (sk_typ & DISKSIM_READ) {
	    skew_ms = 0.295 + 0.115 + 0.000667 * (sk_len < 8 ? 0 : sk_len - 8);
	  } else {
	    skew_ms = 0.295 + 0.115 + 0.000775 * (sk_len < 8 ? 0 : sk_len - 8);
	  }
	}
      }
	  
      time_msec = get_curtime_msec();
      //time_stamp(&time_ts);

      if (completed_time_msec > 0.0
	  && completed_time_msec <= (time_msec + skew_ms /* 0.3, 0.01*/)) {
	send_completion(completed_time_msec, completed_request);

	{
	  double diff = completed_time_msec - get_curtime_msec();
	  if (diff < compl_min) compl_min = diff;
	  if (diff > compl_max) compl_max = diff;
	  if (diff < 0.0) compl_num_neg++;
	  compl_total += diff;
	  compl_num++;
	}

	//completed_time_ts = 0;
	completed_request = NULL;
	completed_time_msec = 0.0;
	yield_ok = 0;
      }

      if (next_simtime_msec > 0.0
	  && next_simtime_msec <= (time_msec + skew_ms /* 0.3, 0.02*/)) {
	//next_callback_ts = 0;
	prev_simtime_msec = next_simtime_msec;
	next_simtime_msec = 0.0;
	callback_fn(disksim_interface, prev_simtime_msec, 0);
	yield_ok = 0;
      }

      if (!yield_ok) continue;
    }

    sched_yield();
  }

  return (0);  /* GRONK */
}


/************************************************************************
 * Main routines
 ************************************************************************/

static void
process_args (int argc, char **argv)
{
  char *program = argv[0];
  int ch;

  while ((ch = getopt(argc, argv, "df:k:")) != -1) {
    switch (ch) {
    case 'd':
      debug=1;
      break;
    case 'f':
      calibration_prefix = optarg;
      break;
    case 'k':
      skew = optarg;
      log("skew set to %c\n", *skew);
      break;
    default:
      goto usage;
    }
  }
  argc -= optind;
  argv += optind;

  if (argc != 0) {
  usage:
    printf("\tUsage: %s [-d] [-f calibration_prefix]\n"
	   "\t\t-d: debug\n"
	   , program);
    assert(0);
  }
}


int
main (int argc, char **argv)
{
  int ret=0;

  process_args(argc, argv);

  open_calibration_times_file();

  start_simulation();

  memulator_attach();

  ret = main_loop();

  stop_simulation();

  close_calibration_times_file();

  log("compl_min = %f\n", compl_min);
  log("compl_max = %f\n", compl_max);
  log("compl_num = %d\n", compl_num);
  log("compl_avg = %f\n", compl_total / compl_num);
  log("compl_num_neg = %d\n", compl_num_neg);

  return (ret);
}
