/*
 * The "Memulator" timing-accurate storage emulation environment
 * Authors: John Linwood Griffin, John Bucy, Greg Ganger
 *
 * Copyright (c) of Carnegie Mellon University, 2001-2004.
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 *
 */

#define MEMULATOR_EVENT_TARGET_EMULATOR_ARRIVE	 (0x31)
#define MEMULATOR_EVENT_TARGET_EMULATOR_COMPLETE (0x32)
#define MEMULATOR_EVENT_TARGET_SIMULATOR_TIME    (0x40)

/*
 * SCSI Disk Emulator
 * 
 * Copyright (c) 2002 Nate Lawson.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification, immediately at the beginning of the file.
 * 2. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>
#include <errno.h>
//#include <err.h>
#include <fcntl.h>
#include <signal.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
//#include <sysexits.h>
#include <unistd.h>
#include <sys/uio.h>
//#include <aio.h>
#include <assert.h>
//#include <sys/stat.h>
#include <sys/queue.h>
#include <sys/event.h>
#include <sys/param.h>  /* DFLTPHYS */
#include <cam/cam_queue.h>
#include <cam/scsi/scsi_all.h>
#include <cam/scsi/scsi_targetio.h>
#include <cam/scsi/scsi_message.h>
#include <sys/socket.h>		/* For IPC */
#include <sys/un.h>		/* For IPC */
#include <sys/shm.h>
#include <sys/sysctl.h>
#include <time.h>
#include <sched.h>

#include "memulator.h"

#define USE_DISKSIM
#ifdef USE_DISKSIM
#include "disksim/src/disksim_interface.h"
#endif


#define MAX_INITIATORS (256)

/************************************************************************
 * Global definitions and function prototypes
 ************************************************************************/

int debug=0;
char *store_filename=NULL;  /* File to use for backing store open */
char *store_memory_size=NULL;  /* Size of memory to use for store */
struct ioc_enable_lun ioc_enlun;  /* Holds bus:target:lun information */
int targ_fd=-1;  /* Descriptor for communication with device driver */
u_int16_t req_flags=0;  /* User-requested HBA configuration */
u_int16_t sim_flags=0;  /* Capabilities of the SCSI interface module */
u_int32_t volume_size_blocks=1024;
int kq_fd=-1;  /* Descriptor for kernel event queue */
int listen_sock=-1;  /* Descriptor for new IPC connections */
int store_sock=-1;  /* Descriptor for IPC connection with store module */
int logic_sock=-1;  /* Descriptor for IPC connection with logic module */
char *times_filename=NULL;  /* File to use for times output */
int times_fd=-1;
int times_number=1;

int clock_enabled = 0;  /* Set when the user specifies to use a clock */
double skew_ms=0.0;  /* Completion time skew for calibration purposes */
double lookahead_ms = 0.0;  /* Current-time lookahead (ms) */
double busywait_ms = 0.0;  /* How far in advance of completion to busy-wait */

char *calibration_prefix=NULL;  /* Prefix to use for calibration times out */
int calibration_fd=-1;
char *memory_preload_file = NULL;  /* Pass to store module */
char *trace_dirname = NULL;  /* Pass to store module */


typedef struct box {
  TAILQ_ENTRY(box) tqe;
  union {
    union ccb *ccb;
    struct request *req;
    struct ccb_scsiio *ctio;
    u_int8_t *buf;
  } u;
} box_t;
TAILQ_HEAD(, box) free_boxes = TAILQ_HEAD_INITIALIZER(free_boxes);

struct buf_descr {
  TAILQ_ENTRY(buf_descr) tqe;
  int serial;
  int size;
#define LOCK_DATAXFER	(0x00000002)
#define LOCK_STORE	(0x00000004)
#define LOCK_LOGIC	(0x00000008)
  int status;
  int shmid;
  u_int8_t *shmbuf;
};
TAILQ_HEAD(, buf_descr) active_bufs = TAILQ_HEAD_INITIALIZER(active_bufs);
TAILQ_HEAD(, buf_descr) free_bufs = TAILQ_HEAD_INITIALIZER(free_bufs);
int num_free_bufs=0;
int min_free_bufs=16;

typedef struct request {
  /* Lists of active and free requests are maintained using this pointer */
  TAILQ_ENTRY(request) tqe;

  /* Unique identifier for this request */
  int serial;

  /* This is the ATIO (Accept Target I/O) we received from the device
   * driver to start this request.  The CDB (Command Descriptor Block)
   * pointer is for convenience, and points into the ATIO structure. */
  struct ccb_accept_tio *atio;
  u_int8_t *cdb;

  /* These keep track of the current state of the request, and what
   * needs to happen before the request can be completed. */
#define TYPE_READ		(1)
#define TYPE_WRITE		(2)
#define TYPE_OTHER		(3)
#define TYPE_INTERPOSE		(4)
#define TYPE_ADMIN_READ		(5)
#define TYPE_ADMIN_WRITE	(6)
#define TYPE_INITIALIZER	(7)
  u_int16_t type;

#define PENDING_COMPLETION	(0x0001)  /* Request completion sent */
#define PENDING_DATAXFER	(0x0002)  /* Data transfer not complete */
#define PENDING_STORE		(0x0004)  /* Store module not finished */
#define PENDING_LOGIC		(0x0008)  /* Logic module not finished */
#define PENDING_CLOCK		(0x0010)  /* Clock module not finished */
#define SENT_COMPLETION		(0x0100)
#define STARTED_DATAXFER	(0x0200)
#define STARTED_STORE		(0x0400)
#define STARTED_LOGIC		(0x0800)
#define STARTED_CLOCK		(0x1000)
  u_int16_t status;

  /* Starting sector and number of sectors in this request. */
  unsigned int offset_blocks;
  unsigned int length_blocks;

  /* CTIO management: to transfer data between the device driver and us,
   * we use CTIOs (Continue Target I/O).  Each CTIO is limited to
   * transferring DFLTPHYS bytes of data, so we may need more than one
   * to satisfy a request. */
  TAILQ_HEAD(, box) ctios_pending;
  unsigned int ctio_flags;
  int dataxfer_bytes_remaining;

  /* Pointer to the buffer if one is allocated for this request */
  u_int8_t *buffer;

  /* Tailq for clock module support: requests are added to the
   * clock_complete tailq if they have finished in the simulator but it is
   * not quite yet time to send the completion notification. */
  TAILQ_ENTRY(request) clock_complete_tqe;
  double clock_completion_simtime;

  /* Time stamps taken at various points during a request's lifetime. */
  u_int64_t ts_arrival;
  u_int64_t ts_dataxfer_begin;
  u_int64_t ts_dataxfer_end;
  u_int64_t ts_logic_send;
  u_int64_t ts_logic_begin;
  u_int64_t ts_logic_end;
  u_int64_t ts_logic_recv;
  u_int64_t ts_store_send;
  u_int64_t ts_store_begin;
  u_int64_t ts_store_end;
  u_int64_t ts_store_recv;
  u_int64_t ts_completion;
} request_t;

TAILQ_HEAD(, request) active_reqs = TAILQ_HEAD_INITIALIZER(active_reqs);
int num_active_reqs=0;
TAILQ_HEAD(, request) free_reqs = TAILQ_HEAD_INITIALIZER(free_reqs);
int num_free_reqs=0;  /* Number of elements in the free_reqs list */
const int min_free_reqs=32;  /* Low-water mark for free_reqs list */


void try_to_complete_request(request_t *);

/************************************************************************
 * Auxiliary functions
 ************************************************************************/

static inline u_int
min (u_int a, u_int b)
{
  return (a < b ? a : b);
}

#define log(a,b...) \
   do { fprintf(stdout, "%s: %s: " a, "memulator", __func__, b); } while (0)

void *
xmalloc (size_t size)
{
  register void *p = malloc(size);
  if (!p) {
    log("malloc %d: %s\n", size, strerror(errno));
    assert(0);
  }
  return (p);
}

void *
xcalloc (size_t count, size_t size)
{
  register void *p = calloc(count, size);
  if (!p) {
    log("malloc %d: %s\n", size, strerror(errno));
    assert(0);
  }
  return (p);
}

#define set_status(x,value)		\
  do { 					\
    assert(!(x->status & value));	\
    x->status |= value;			\
  } while (0)

#define clear_status(x,value)		\
  do { 					\
    assert(x->status & value);		\
    x->status &= ~value;		\
  } while (0)

/************************************************************************
 * Timing support
 ************************************************************************/

inline double
time_stamp (void)
{
  u_int64_t d;
  __asm__ __volatile__ ("rdtsc" : "=&A" (d));
  return d;
}

void
print_request_times (request_t *req)
{
  int rc;
  int bytes_written=0;
  int bytes_remaining=4096;
  u_char *index;
  u_char *type;
  u_char output[bytes_remaining];

  if (times_fd==-1) return;

  index = output;

  switch (req->type) {
  case TYPE_READ:        type = "read"; break;
  case TYPE_WRITE:       type = "write"; break;
  case TYPE_OTHER:       type = "other"; break;
  case TYPE_INTERPOSE:   type = "interpose_read"; break;
  case TYPE_ADMIN_READ:  type = "admin_read"; break;
  case TYPE_ADMIN_WRITE: type = "admin_write"; break;
  default:
    log("unknown type: %x\n", req->type);
    type = "unknown";
  }

  rc = snprintf(index, bytes_remaining, "%d:%d arrival_%s %llu %d\n",
		req->offset_blocks, req->length_blocks, type,
		req->ts_arrival, req->serial);
  assert(rc >= 0);
  bytes_written += rc;
  bytes_remaining -= rc;
  assert(bytes_remaining >= 0);
  index += rc;

#define PRINT_TIMES_ENTRY(WHICH, NAME)					\
  if (req->WHICH) {							\
    rc = snprintf(index, bytes_remaining, "%d:%d %s %llu %d\n",		\
		  req->offset_blocks, req->length_blocks, NAME,		\
		  req->WHICH, req->serial);				\
    assert(rc >= 0);							\
    bytes_written += rc;						\
    bytes_remaining -= rc;						\
    assert(bytes_remaining >= 0);					\
    index += rc;							\
  }

  PRINT_TIMES_ENTRY(ts_dataxfer_begin,	"dataxfer_begin");
  PRINT_TIMES_ENTRY(ts_dataxfer_end,	"dataxfer_end");
  PRINT_TIMES_ENTRY(ts_logic_send,	"logic_send");
  PRINT_TIMES_ENTRY(ts_logic_begin,	"logic_begin");
  PRINT_TIMES_ENTRY(ts_logic_end,	"logic_end");
  PRINT_TIMES_ENTRY(ts_logic_recv,	"logic_recv");
  PRINT_TIMES_ENTRY(ts_store_send,	"store_send");
  PRINT_TIMES_ENTRY(ts_store_begin,	"store_begin");
  PRINT_TIMES_ENTRY(ts_store_end,	"store_end");
  PRINT_TIMES_ENTRY(ts_store_recv,	"store_recv");

  rc = snprintf(index, bytes_remaining, "%d:%d completion_%s %llu %d\n",
		req->offset_blocks, req->length_blocks, type,
		req->ts_completion, req->serial);
  assert(rc >= 0);
  bytes_written += rc;
  bytes_remaining -= rc;
  assert(bytes_remaining >= 0);
  index += rc;

  if (write(times_fd, output, bytes_written) == -1) {
    log("write: %s\n", strerror(errno));
  }
}


void
open_calibration_times_file (void)
{
  const int filename_maxlen=80;
  char filename[filename_maxlen];

  if (!calibration_prefix) return;

  if ((strlen(calibration_prefix) + 4) > filename_maxlen) {
    log("please increase filename_maxlen (%d)\n", filename_maxlen);
    assert(0);
  }

  snprintf(filename, filename_maxlen,
	   "%s.mem",
	   calibration_prefix);

  calibration_fd = open(filename, O_CREAT|O_WRONLY, 0644);
  if (calibration_fd == -1) {
    log("unable to open %s: %s\n", filename, strerror(errno));
    assert(0);
  }
}


void
close_calibration_times_file (void)
{
  if (calibration_fd == -1) return;

  if (close(calibration_fd) == -1) {
    log("close: %s\n", strerror(errno));
  }
  calibration_fd = -1;
}


void
calibration_time_stamp_arrive (request_t *req)
{
  const int buflen = 160;
  unsigned char buf[buflen];
  int bytes;

  if (calibration_fd == -1) return;

  bytes = snprintf(buf, buflen,
		   "%d\t%d\t%x\t%x\t%lld\t%d\n",
		   req->offset_blocks,
		   req->length_blocks,
		   req->cdb[0],
		   MEMULATOR_EVENT_TARGET_EMULATOR_ARRIVE,
		   req->ts_arrival,
		   req->serial);

  write(calibration_fd, buf, bytes);
}


void
calibration_time_stamp_complete (request_t *req)
{
  const int buflen = 160;
  unsigned char buf[buflen];
  int bytes;

  if (calibration_fd == -1) return;

  bytes = snprintf(buf, buflen,
		   "%d\t%d\t%x\t%x\t%lld\t%d\n",
		   req->offset_blocks,
		   req->length_blocks,
		   req->cdb[0],
		   MEMULATOR_EVENT_TARGET_EMULATOR_COMPLETE,
		   req->ts_completion,
		   req->serial);

  write(calibration_fd, buf, bytes);
}


void
open_times_file (void)
{
  const int filename_maxlen=80;
  char filename[filename_maxlen];
  time_t now;
  u_int64_t timestamp;
  int len;
  const int maxlen=512;
  char buf[maxlen];

  if (!times_filename) return;

  if ((strlen(times_filename) + 4) > filename_maxlen) {
    log("please increase filename_maxlen (%d)\n", filename_maxlen);
    assert(0);
  }

  snprintf(filename, filename_maxlen,
	   "%s.%3.3d",
	   times_filename, times_number++);

  times_fd = open(filename, O_CREAT|O_EXCL|O_WRONLY, 0644);
  if (times_fd == -1) {
    log("unable to open %s: %s\n", filename, strerror(errno));
    assert(0);
  }

  timestamp = time_stamp();
  time(&now);

  len = snprintf(buf, maxlen,
		 "Startup: %lld (%.24s)\n",
		 timestamp, ctime(&now));

  if (len == -1) {
    log("snprintf: %s\n", strerror(errno));
    return;
  }

  if (write(times_fd, buf, len) == -1) {
    log("write: %s\n", strerror(errno));
  }
}


void
close_times_file (void)
{
  time_t now;
  u_int64_t timestamp;
  int len;
  const int maxlen=512;
  char buf[maxlen];

  if (times_fd == -1) return;

  timestamp = time_stamp();
  time(&now);

  len = snprintf(buf, maxlen,
		 "Shutdown: %lld (%.24s)\n",
		 timestamp, ctime(&now));

  if (len == -1) {
    log("snprintf: %s\n", strerror(errno));
    return;
  }

  if (write(times_fd, buf, len) == -1) {
    log("write: %s\n", strerror(errno));
  }

  if (close(times_fd) == -1) {
    log("close: %s\n", strerror(errno));
  }
}


/************************************************************************
 * Debug support
 ************************************************************************/

void
cdb_debug(u_int8_t *cdb, const char *msg, ...)
{
  char msg_buf[512];
  int len;
  va_list ap;

  va_start(ap, msg);
  vsnprintf(msg_buf, sizeof(msg_buf), msg, ap);
  va_end(ap);
  len = strlen(msg_buf);
  scsi_cdb_string(cdb, msg_buf + len, sizeof(msg_buf) - len);
  printf("%s\n", msg_buf);
  fflush(stdout);
}


void
print_sector(u_int8_t *p)
{
  int i;
  printf("memulator: ");
  for (i=0;i<512;i++) {
    printf("%#x ", p[i]);
    if (!((i+1)%16)) printf("\nmemulator: ");
  }
  printf("--------------------------------------------------------------\n");
}

/************************************************************************
 * Box
 ************************************************************************/

box_t *
get_box (void)
{
  box_t *box;

  if (!TAILQ_EMPTY(&free_boxes)) {
    box = TAILQ_FIRST(&free_boxes);
    TAILQ_REMOVE(&free_boxes, box, tqe);
  } 
  else {
    box = xmalloc(sizeof(box_t));
  }

  memset(box, 0, sizeof(box_t));
  return (box);
}


void
release_box (box_t *box)
{
  TAILQ_INSERT_TAIL(&free_boxes, box, tqe);
}


/************************************************************************
 * CCB freelist management functions
 ************************************************************************/

TAILQ_HEAD(, box) free_ccbs = TAILQ_HEAD_INITIALIZER(free_ccbs);

int num_free_ccbs=0;  /* Number of elements in the free_ccbs list */
const int min_free_ccbs=32;  /* Low-water mark for the free_ccbs list */

union ccb *
get_free_ccb (void)
{
  box_t *box;
  union ccb *ccb;

  /* First check if we've reached the low-water mark for free CCBs */

  if (num_free_ccbs < min_free_ccbs) {
    int i;
    for (i=0; i<min_free_ccbs; i++) {
      box_t *new_box;
      union ccb *new_ccb;

      /* Get a container for the CCB */

      new_box = get_box();

      /* Allocate space for the new CCB */

      new_ccb = xcalloc(1, sizeof(union ccb));

      /* Put the new ccb in the box, and add to the free_ccbs list. */

      new_box->u.ccb = new_ccb;
      TAILQ_INSERT_TAIL(&free_ccbs, new_box, tqe);
      num_free_ccbs++;
    }
  }

  assert(!TAILQ_EMPTY(&free_ccbs));

  /* Grab the next available entry from the list of free CCBs. */

  box = TAILQ_FIRST(&free_ccbs);
  TAILQ_REMOVE(&free_ccbs, box, tqe);
  num_free_ccbs--;

  /* Place the container on the list of empty containers. */

  ccb = box->u.ccb;
  box->u.ccb = NULL;
  release_box(box);

  /* Zero-fill the new CCB. */

  memset(ccb, 0, sizeof(union ccb));
  return (ccb);
}


void
release_ccb (union ccb *ccb)
{
  box_t *box;

  box = get_box();
  box->u.ccb = ccb;

  TAILQ_INSERT_TAIL(&free_ccbs, box, tqe);
  num_free_ccbs++;
}


void
send_ccb (union ccb *ccb)
{
  ccb->ccb_h.pinfo.priority = 1;

  if (debug) log("ccb %p\n", ccb);

  /* GRONK: The example code checks for XPT_FC_IS_QUEUED(ccb) here:
   * 
   * if (XPT_FC_IS_QUEUED(ccb)) {
   *   TAILQ_INSERT_TAIL(&pending_queue, &ccb->ccb_h, periph_links.tqe);
   * }
   */

  if (write(targ_fd, &ccb, sizeof(ccb)) != sizeof(ccb)) {
    log("write: %s\n", strerror(errno));
    assert(0);
  }
}


/************************************************************************
 * Buffer management functions
 ************************************************************************/

u_int8_t *
allocate_buffer (int serial, int size)
{
  struct buf_descr *buf;

  /* First check if we've reached the low-water mark for free buf_descrs */

  if (num_free_bufs < min_free_bufs) {
    int i;
    for (i=0; i<min_free_bufs; i++) {
      struct buf_descr *new_buf;

      /* Allocate space for the new buf_descr */

      new_buf = xcalloc(1, sizeof(struct buf_descr));

      /* Add the new buf_descr to the free_bufs list */

      TAILQ_INSERT_TAIL(&free_bufs, new_buf, tqe);
      num_free_bufs++;
    }
  }

  assert(!TAILQ_EMPTY(&free_bufs));

  /* Grab the next available buf_descr from the list of free bufs. */

  buf = TAILQ_FIRST(&free_bufs);
  TAILQ_REMOVE(&free_bufs, buf, tqe);
  num_free_bufs--;

  /* Initialize the values for the new buf_descr. */

  memset(buf, 0, sizeof(struct buf_descr));

  buf->serial = serial;
  buf->size = size;

  if ((buf->shmid = shmget(buf->serial, buf->size, IPC_CREAT | 0666)) == -1) {
    log("shmget %d (%d): %s\n", buf->serial, buf->size, strerror(errno));
    // assert(0);
  }

  if ((buf->shmbuf = shmat(buf->shmid, NULL, 0)) == (void *)-1) {
    log("shmat: %s\n", strerror(errno));
    // assert(0);
  }

  TAILQ_INSERT_TAIL(&active_bufs, buf, tqe);

  return (buf->shmbuf);
}


void
buffer_lock (int serial, int whom)
{
  struct buf_descr *buf;

  buf = TAILQ_FIRST(&active_bufs);
  while (buf && buf->serial != serial) {
    buf = TAILQ_NEXT(buf, tqe);
  }
  assert(buf);

  assert(!(buf->status & whom));
  set_status(buf, whom);
}


void
buffer_unlock (int serial, int whom)
{
  struct buf_descr *buf;

  buf = TAILQ_FIRST(&active_bufs);
  while (buf && buf->serial != serial) {
    buf = TAILQ_NEXT(buf, tqe);
  }
  assert(buf);

  assert(buf->status & whom);
  clear_status(buf, whom);

  if (!buf->status) {
    if (shmdt(buf->shmbuf) == -1) {
      log("shmdt: %s\n", strerror(errno));
    }

    if (shmctl(buf->shmid, IPC_RMID, NULL) == -1) {
      log("shmctl: %s\n", strerror(errno));
    }

    TAILQ_REMOVE(&active_bufs, buf, tqe);

    TAILQ_INSERT_TAIL(&free_bufs, buf, tqe);
    num_free_bufs++;
  }
}


/************************************************************************
 * Request freelist management functions
 ************************************************************************/

request_t *
get_free_request (void)
{
  static int next_serial=1;
  request_t *req;

  /* First check if we've reached the low-water mark for free requests */

  if (num_free_reqs < min_free_reqs) {
    int i;
    for (i=0; i<min_free_reqs; i++) {
      request_t *new_req;

      /* Allocate space for the new request */

      new_req = xcalloc(1, sizeof(request_t));

      /* Add the new request to the free_reqs list */

      TAILQ_INSERT_TAIL(&free_reqs, new_req, tqe);
      num_free_reqs++;
    }
  }

  assert(!TAILQ_EMPTY(&free_reqs));

  /* Grab the next available entry from the list of free reqs. */

  req = TAILQ_FIRST(&free_reqs);
  TAILQ_REMOVE(&free_reqs, req, tqe);
  num_free_reqs--;

  /* Zero-fill the new request. */

  memset(req, 0, sizeof(request_t));
  req->serial = next_serial++;
  TAILQ_INIT(&req->ctios_pending);

  /* Place the container on the list of active requests */

  TAILQ_INSERT_TAIL(&active_reqs, req, tqe);
  num_active_reqs++;

  return (req);
}



void
release_request (request_t *req)
{
  /* Remove request from list of active requests */

  TAILQ_REMOVE(&active_reqs, req, tqe);
  num_active_reqs--;

  if (debug) log("releasing %d\n", req->serial);

  /* Send replacement ATIO down to driver */

  if (req->atio) {
    struct ccb_accept_tio *atio;

    release_ccb((union ccb *)req->atio);
    req->atio = NULL;

    atio = (struct ccb_accept_tio *)get_free_ccb();
    atio->ccb_h.func_code = XPT_ACCEPT_TARGET_IO;
    send_ccb((union ccb *)atio);
  }

  /* Add request to list of free requests */

  TAILQ_INSERT_TAIL(&free_reqs, req, tqe);
  num_free_reqs++;
}
  

void
attach_ctio_to_request (struct ccb_scsiio *ctio, request_t *req)
{
  box_t *box;

  if (debug) log("Attaching %p to %d\n", ctio, req->serial);

  box = get_box();
  box->u.ctio = ctio;

  TAILQ_INSERT_TAIL(&req->ctios_pending, box, tqe);
}


void
start_dataxfer (request_t *req)
{
  box_t *b;

  req->ts_dataxfer_begin = time_stamp();

  set_status(req, STARTED_DATAXFER);
  clear_status(req, PENDING_DATAXFER);

  TAILQ_FOREACH(b, &req->ctios_pending, tqe) {
    struct ccb_scsiio *ctio = b->u.ctio;
    send_ccb((union ccb *)ctio);
  }
}


request_t *
get_request_by_serial (int serial)
{
  request_t *req;

  TAILQ_FOREACH(req, &active_reqs, tqe) {
    if (req->serial == serial) {
      return (req);
    }
  }

  log("Unable to locate request %d\n", serial);
  return (NULL);
}


request_t *
get_request_by_ctio (struct ccb_scsiio *ctio)
{
  request_t *req;
  box_t *ctiobox;

  TAILQ_FOREACH(req, &active_reqs, tqe) {
    ctiobox = TAILQ_FIRST(&req->ctios_pending);
    while (ctiobox) {
      if (ctiobox->u.ctio==ctio) {
	return (req);
      }
      ctiobox = TAILQ_NEXT(ctiobox, tqe);
    }
  }

  log("Unable to locate CTIO %p\n", ctio);
  assert(0);
  exit(1);
}


/************************************************************************
 * Interprocess communication
 ************************************************************************/

int
send_ipc (int sock, void *buf, size_t len)
{
  u_int8_t *msg;
  int bytes = 0;	/* Total bytes sent */
  int rc;

  if (debug) log("%s\n","send_ipc");

  msg = buf;

  while (bytes < len) {
    rc = send(sock, msg + bytes, len - bytes, 0);
    if (rc == -1) {
      log("send: %s\n", strerror(errno));
      return (-1);
    }
    else if (rc == 0) {
      log("send: %s\n", "returned zero");
      goto done;
    }
    else {
      bytes += rc;
    }
  }

 done:
  if (bytes != len) {
    log("bad transfer (%d != %d)\n", bytes, len);
  }

  return (bytes);
}


int
recv_ipc (int sock, void *buf, size_t len)
{
  u_int8_t *msg;
  int bytes = 0;	/* Total bytes received */
  int rc;

  if (debug) log("%s\n","recv_ipc");

  msg = buf;

  while (bytes < len) {
    rc = recv(sock, buf + bytes, len - bytes, 0);
    if (rc == -1) {
      log("recv: %s\n", strerror(errno));
      return (-1);
    }
    else if (rc == 0) {
      log("recv: %s\n", "returned zero");
      goto done;
    }
    else {
      bytes += rc;
    }
  }

 done:
  if (bytes != len) {
    log("bad transfer (%d != %d)\n", bytes, len);
  }

  return (bytes);
}


void
enable_ipc (void)
{
  struct sockaddr_un sun;

  assert(listen_sock == -1);

  /* socket */

  if ((listen_sock = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) {
    log("socket: %s\n", strerror(errno));
    assert(0);
  }

  /* bind */

  unlink(MEMULATOR_SOCKET_FILENAME);
  memset(&sun, 0, sizeof(struct sockaddr_un));
  sun.sun_family = AF_UNIX;
  strncpy(sun.sun_path, MEMULATOR_SOCKET_FILENAME, 104);
  if (bind(listen_sock, (struct sockaddr *)&sun, SUN_LEN(&sun)) == -1) {
    log("bind: %s\n", strerror(errno));
    assert(0);
  }

  /* listen */

  if (listen(listen_sock, 10) == -1) {
    log("listen: %s\n", strerror(errno));
    assert(0);
  }
}


void
disable_ipc (void)
{
  assert(listen_sock != -1);

  if (close(listen_sock) == -1) {
    log("close: %s\n", strerror(errno));
  }
  listen_sock = -1;

  unlink(MEMULATOR_SOCKET_FILENAME);
}


void
start_store (void)
{
  const int max_args=20;
  char *argv[max_args];
  int n=0;
  pid_t pid;

  if (store_filename && store_memory_size) {
    log("%s\n", "Cannot use both -b and -y options!\n");
    exit(1);
  }

  if (!store_filename && !store_memory_size) return;

  argv[n++] = "./store";

  if (store_filename) {
    argv[n++] = "-b";
    argv[n++] = store_filename;
  }
  if (store_memory_size) {  /* note: -b and -y cannot be used together */
    argv[n++] = "-y";
    argv[n++] = store_memory_size;
  }

  if (memory_preload_file) {
    argv[n++] = "-x";
    argv[n++] = memory_preload_file;
  }

  if (trace_dirname) {
    argv[n++] = "-r";
    argv[n++] = trace_dirname;
  }

  if ((pid = fork()) == -1) {
    log("fork: %s\n", strerror(errno));
    assert(0);
  }

  /* GRONK: catch sigchild? */
  if (pid==0) {  /* child */

    if (debug) argv[n++] = "-d";
    argv[n++] = NULL;
    assert(n<=max_args);
    execv(argv[0], argv);
    log("execv: %s\n", strerror(errno));
    assert(0);
  }
}


void
stop_store (void)
{
  ipc_message_t msg_out;

  if (store_sock == -1) return;

  memset(&msg_out, 0, sizeof(ipc_message_t));
  msg_out.magic  = IPC_MAGIC;
  msg_out.source = IPC_SOURCE_MEMULATOR;
  msg_out.type   = IPC_TYPE_DISCONNECT;

  send_ipc(store_sock, &msg_out, sizeof(ipc_message_t));
  if (close(store_sock) == -1) {
    log("close store: %s\n", strerror(errno));
  }
  store_sock=-1;
}


void
set_logic_timeout (int serial)
{
  const int max_events = 1;
  struct kevent events[max_events];
  int n = 0;

  if (0) log("Setting timer for %d\n", serial);

  EV_SET(&events[n++], 
	 serial, /* identifier for this timer */
	 EVFILT_TIMER, 
	 EV_ADD | EV_ENABLE | EV_ONESHOT,
	 0,
	 5000, /* timeout in milliseconds */
	 0);
  assert(n <= max_events);
  if (kevent(kq_fd, events, n, NULL, 0, NULL) == -1) {
    log("kevent: %s\n", strerror(errno));
    assert(0);
  }
}


void
clear_logic_timeout (int serial)
{
  const int max_events = 1;
  struct kevent events[max_events];
  int n = 0;

  if (0) log("Clearing timer for %d\n", serial);

  EV_SET(&events[n++], 
	 serial, /* identifier for this timer */
	 EVFILT_TIMER, 
	 EV_DISABLE | EV_DELETE,
	 0,
	 0,
	 0);
  assert(n <= max_events);
  if (kevent(kq_fd, events, n, NULL, 0, NULL) == -1) {
    log("kevent: %s\n", strerror(errno));
    // assert(0);
  }
}


void
notify_logic (request_t *req)
{
  int rc;
  ipc_message_t msg_out;

  if (debug) log("request %d\n", req->serial);

  if (logic_sock == -1) {
    log("Unable to notify_logic for request %d\n", req->serial);
    return;
  }

  set_status(req, STARTED_LOGIC);
  clear_status(req, PENDING_LOGIC);

  memset(&msg_out, 0, sizeof(ipc_message_t));
  msg_out.magic  = IPC_MAGIC;
  msg_out.source = IPC_SOURCE_MEMULATOR;
  if      (req->type == TYPE_READ)  msg_out.type = IPC_TYPE_READ;
  else if (req->type == TYPE_WRITE) msg_out.type = IPC_TYPE_WRITE;
  else if (req->type == TYPE_INTERPOSE) msg_out.type = IPC_TYPE_INTERPOSE_OK;
  else if (req->type == TYPE_ADMIN_READ)  msg_out.type = IPC_TYPE_ADMIN_READ;
  else if (req->type == TYPE_ADMIN_WRITE) msg_out.type = IPC_TYPE_ADMIN_WRITE;
  else {
    log("Bad type %#x\n", req->type);
    assert(0);
  }
  msg_out.request_offset_blocks = req->offset_blocks;
  msg_out.request_length_blocks = req->length_blocks;
  msg_out.serial = req->serial;

  req->ts_logic_send = time_stamp();

  rc = send_ipc(logic_sock, &msg_out, sizeof(ipc_message_t));
  if (rc != sizeof(ipc_message_t)) {
    log("ERROR: Problem sending request to logic (%d).\n", rc);
  }

  /* Set timer to detect logic module failure */
  if (req->type != TYPE_INTERPOSE) {
    set_logic_timeout(req->serial);
  }
}


void
notify_store (request_t *req)
{
  int rc;
  ipc_message_t msg_out;

  if (debug) log("request %d\n", req->serial);

  assert(store_sock != -1);

  set_status(req, STARTED_STORE);
  clear_status(req, PENDING_STORE);

  memset(&msg_out, 0, sizeof(ipc_message_t));
  msg_out.magic  = IPC_MAGIC;
  msg_out.source = IPC_SOURCE_MEMULATOR;
  if      (req->type == TYPE_READ)  msg_out.type = IPC_TYPE_READ;
  else if (req->type == TYPE_WRITE) msg_out.type = IPC_TYPE_WRITE;
  else if (req->type == TYPE_INTERPOSE) msg_out.type = IPC_TYPE_READ;
  else {
    log("Bad type %#x\n", req->type);
    assert(0);
  }
  msg_out.request_offset_blocks = req->offset_blocks;
  msg_out.request_length_blocks = req->length_blocks;
  msg_out.serial = req->serial;

  req->ts_store_send = time_stamp();

  rc = send_ipc(store_sock, &msg_out, sizeof(ipc_message_t));
  if (rc != sizeof(ipc_message_t)) {
    log("ERROR: Problem sending request to store (%d); system may hang!\n",rc);
  }
  sched_yield();
}


void clock_request_arrive(request_t *req);


void
notify_clock (request_t *req)
{
  if (debug) log("request %d\n", req->serial);

  if (req->offset_blocks == 1090920) {
    log("Falsifying request %d\n", req->serial);
    req->offset_blocks -= 40;
    // clear_status(req, PENDING_CLOCK);
    // return;
  }

  set_status(req, STARTED_CLOCK);
  clear_status(req, PENDING_CLOCK);

  clock_request_arrive(req);
}


void
start_next_phase (struct request *req)
{
  switch (req->type) {

  case TYPE_READ:
    if (debug) log("read %d: %#x\n", req->serial, req->status);
    if ((req->status & PENDING_CLOCK) && !(req->status & PENDING_LOGIC)) {
      notify_clock(req);
    }

    if      (req->status & PENDING_LOGIC)    notify_logic(req);
    else if (req->status & PENDING_STORE)    notify_store(req);
    else if (req->status & PENDING_DATAXFER) start_dataxfer(req);
    else                                     try_to_complete_request(req);
    break;

  case TYPE_WRITE:
    if (debug) log("write %d: %#x\n", req->serial, req->status);
    if ((req->status & PENDING_CLOCK) && !(req->status & PENDING_LOGIC)) {
      notify_clock(req);
    }

    if      (req->status & PENDING_DATAXFER) start_dataxfer(req);
    else if (req->status & PENDING_LOGIC)    notify_logic(req);
    else if (req->status & PENDING_STORE)    notify_store(req);
    else                                     try_to_complete_request(req);
    break;

  case TYPE_OTHER:
    if (debug) log("other %d: %#x\n", req->serial, req->status);
    if      (req->status & PENDING_DATAXFER) start_dataxfer(req);
    else if (req->status & PENDING_LOGIC)    assert(0);
    else if (req->status & PENDING_STORE)    assert(0);
    else if (req->status & PENDING_CLOCK)    assert(0);
    else                                     try_to_complete_request(req);
    break;

  case TYPE_INTERPOSE:
    if (debug) log("interpose %d: %#x\n", req->serial, req->status);
    if      (req->status & PENDING_STORE)    notify_store(req);
    else if (req->status & PENDING_LOGIC)    notify_logic(req);
    else if (req->status & PENDING_DATAXFER) assert(0);
    else if (req->status & PENDING_CLOCK)    assert(0);
    else                                     try_to_complete_request(req);
    break;

  case TYPE_ADMIN_READ:
    if (debug) log("admin read %d: %#x\n", req->serial, req->status);
    if      (req->status & PENDING_LOGIC)    notify_logic(req);
    else if (req->status & PENDING_DATAXFER) start_dataxfer(req);
    else if (req->status & PENDING_STORE)    assert(0);
    else if (req->status & PENDING_CLOCK)    assert(0);
    else                                     try_to_complete_request(req);
    break;

  case TYPE_ADMIN_WRITE:
    if (debug) log("admin write %d: %#x\n", req->serial, req->status);
    if      (req->status & PENDING_DATAXFER) start_dataxfer(req);
    else if (req->status & PENDING_LOGIC)    notify_logic(req);
    else if (req->status & PENDING_STORE)    assert(0);
    else if (req->status & PENDING_CLOCK)    assert(0);
    else                                     try_to_complete_request(req);
    break;

  default:
    log("unknown type: %d\n", req->type);
    assert(0);
  }
}


void
handle_new_ipc_connection (void)
{
  int rc;
  int new_sock;
  ipc_message_t msg_in;
  const int max_events=1;
  struct kevent events[max_events];
  int n=0;

  if ((new_sock = accept(listen_sock, NULL, NULL)) == -1) {
    log("accept: %s\n", strerror(errno));
    assert(0);
  }

  /* Receive the explanation of who is connecting to us */

  memset(&msg_in, 0, sizeof(ipc_message_t));
  rc = recv_ipc(new_sock, &msg_in, sizeof(ipc_message_t));
  if (rc != sizeof(ipc_message_t)) {
    log("bad length %d\n", rc);
    close(new_sock);
    return;
  }

  if (msg_in.magic != IPC_MAGIC) {
    log("bad magic %#x\n", msg_in.magic);
    close(new_sock);
    return;
  }

  if (msg_in.type != IPC_TYPE_CONNECT) {
    log("bad type %#x\n", msg_in.type);
    close(new_sock);
    return;
  }

  switch (msg_in.source) {

  case IPC_SOURCE_STORE:
    if (store_sock != -1) {
      log("%s\n","multiple connections from store!");
      close(new_sock);
      return;
    }

    store_sock = new_sock;
    EV_SET(&events[n++], store_sock, EVFILT_READ, EV_ADD, 0,0,0);
    assert(n<=max_events);
    if (kevent(kq_fd, events, n, NULL, 0, NULL) == -1) {
      log("kevent: %s\n", strerror(errno));
      assert(0);
    }
    return;

  case IPC_SOURCE_LOGIC:
    if (logic_sock != -1) {
      log("%s\n", "multiple connections from logic!");
      close(new_sock);
      return;
    }

    logic_sock = new_sock;
    EV_SET(&events[n++], logic_sock, EVFILT_READ, EV_ADD, 0,0,0);
    assert(n<=max_events);
    if (kevent(kq_fd, events, n, NULL, 0, NULL) == -1) {
      log("kevent: %s\n", strerror(errno));
      assert(0);
    }
    return;

  default:
    log("unknown source %d\n", msg_in.source);
    close(new_sock);
    return;
  }
}


/************************************************************************
 * Request CCB (ATIO/CTIO) processing
 ************************************************************************/

void
setup_ctio (struct ccb_scsiio *ctio, request_t *req)
{
  ctio->ccb_h.func_code   = XPT_CONT_TARGET_IO;
  ctio->ccb_h.retry_count = 2;
  ctio->ccb_h.timeout     = CAM_TIME_INFINITY;

  ctio->ccb_h.flags = req->ctio_flags;
  ctio->tag_id      = req->atio->tag_id;
  ctio->init_id     = req->atio->init_id;
}


void
send_request_complete (request_t *req)
{
  struct ccb_scsiio *ctio;

  if (debug) log("request %d\n", req->serial);

  ctio = (struct ccb_scsiio *)get_free_ccb();
  setup_ctio(ctio, req);

  ctio->ccb_h.flags &= ~CAM_DIR_MASK;
  ctio->ccb_h.flags |= CAM_DIR_NONE | CAM_SEND_STATUS;
  ctio->scsi_status = SCSI_STATUS_OK;

  attach_ctio_to_request(ctio, req);

  /* Send response */

  send_ccb((union ccb *)ctio);
  sched_yield();
}


void
send_check_condition (request_t *req,
		      u_int8_t flags, u_int8_t asc, u_int8_t ascq)
{
  struct ccb_scsiio *ctio;
  struct scsi_sense_data *sense_data;

  ctio = (struct ccb_scsiio *)get_free_ccb();
  setup_ctio(ctio, req);

  /* Fill in sense response data */

  sense_data = &ctio->sense_data;

  sense_data->error_code = SSD_CURRENT_ERROR;
  sense_data->flags = flags;
  sense_data->add_sense_code = asc;
  sense_data->add_sense_code_qual = ascq;
  sense_data->extra_len = offsetof(struct scsi_sense_data, sense_key_spec[2])
    - offsetof(struct scsi_sense_data, extra_len);

  ctio->sense_len = sizeof(struct scsi_sense_data);
  ctio->dxfer_len = 0;

  ctio->ccb_h.flags &= ~CAM_DIR_MASK;
  ctio->ccb_h.flags |= CAM_DIR_NONE | CAM_SEND_SENSE | CAM_SEND_STATUS;
  ctio->scsi_status = SCSI_STATUS_CHECK_COND;

  /* Activate request */

  req->type = TYPE_OTHER;

  set_status(req, PENDING_DATAXFER);
  attach_ctio_to_request(ctio, req);

  /* Send response */

  start_next_phase(req);
}


void
do_nothing (request_t *req)
{
  struct ccb_scsiio *ctio;

  ctio = (struct ccb_scsiio *)get_free_ccb();
  setup_ctio(ctio, req);

  ctio->ccb_h.flags &= ~CAM_DIR_MASK;
  ctio->ccb_h.flags |= CAM_DIR_NONE;

  /* Activate request */

  req->type = TYPE_OTHER;

  set_status(req, PENDING_COMPLETION);
  set_status(req, PENDING_DATAXFER);
  attach_ctio_to_request(ctio, req);

  /* Send response */

  start_next_phase(req);
}


void
do_inquiry (request_t *req)
{
  struct scsi_inquiry *inq;
  struct scsi_inquiry_data *inq_data;
  struct ccb_scsiio *ctio;

  /* Check for illegal request flags */

  inq = (struct scsi_inquiry *)req->cdb;

  if ((inq->byte2 & SI_EVPD) != 0) {
    send_check_condition(req, SSD_KEY_ILLEGAL_REQUEST, 0x24, 0x00);
    /* GRONK: setup sense */
    return;
  }
  if (inq->page_code != 0) {
     send_check_condition(req, SSD_KEY_ILLEGAL_REQUEST, 0x24, 0x00);
    /* GRONK: setup sense */
    return;
  }

  /* Fill in inquiry response data */

  req->buffer = 
    allocate_buffer(req->serial, sizeof(struct scsi_inquiry_data));
  inq_data = (struct scsi_inquiry_data *)req->buffer;

  inq_data->device  = T_DIRECT | (SID_QUAL_LU_CONNECTED << 5);
  inq_data->version = 3; /* was 2 */

#define SID_Addr16	0x0100
  if ((req_flags & SID_Addr16) != 0) {
    sim_flags |= SID_Addr16;
    log("%s\n","Not sure SIM supports Addr16 but enabling it anyway");
  }

  scsi_ulto2b((req_flags & sim_flags), &inq_data->reserved[1]);
  // scsi_ulto2b((sim_flags & ~SID_CmdQue), &inq_data->reserved[1]);

  inq_data->response_format = 2;  /* SCSI2 Inquiry Format */
  inq_data->additional_length = SHORT_INQUIRY_LENGTH -
    offsetof(struct scsi_inquiry_data, additional_length);

  strncpy(inq_data->vendor, "Carnegie", SID_VENDOR_SIZE);
  strncpy(inq_data->product, "Mellon Memulator", SID_PRODUCT_SIZE);
  strncpy(inq_data->revision, "2004", SID_REVISION_SIZE);

  req->dataxfer_bytes_remaining =
    min(inq_data->additional_length + 4, SCSI_CDB6_LEN(inq->length));
  req->ctio_flags |= CAM_DIR_IN;

  /* Prepare CTIO to transmit inquiry response data */
 
  ctio = (struct ccb_scsiio *)get_free_ccb();
  setup_ctio(ctio, req);
  ctio->data_ptr = (u_int8_t *)inq_data;
  ctio->dxfer_len = req->dataxfer_bytes_remaining;

  /* Activate request */

  req->type = TYPE_OTHER;

  set_status(req, PENDING_COMPLETION);
  set_status(req, PENDING_DATAXFER);
  buffer_lock(req->serial, LOCK_DATAXFER);

  attach_ctio_to_request(ctio, req);

  /* Send response */

  start_next_phase(req);
}


void
do_read_capacity (request_t *req)
{
  struct scsi_read_capacity_data *cap_data;
  struct ccb_scsiio *ctio;

  req->buffer = 
    allocate_buffer(req->serial, sizeof(struct scsi_read_capacity_data));
  cap_data = (struct scsi_read_capacity_data *)req->buffer;

  scsi_ulto4b(volume_size_blocks-1, cap_data->addr);
  scsi_ulto4b(512, cap_data->length);

  req->dataxfer_bytes_remaining = sizeof(struct scsi_read_capacity_data);
  req->ctio_flags |= CAM_DIR_IN;

  /* Prepare CTIO to transmit capacity response data */

  ctio = (struct ccb_scsiio *)get_free_ccb();
  setup_ctio(ctio, req);
  ctio->data_ptr = (u_int8_t *)cap_data;
  ctio->dxfer_len = req->dataxfer_bytes_remaining;

  /* Activate request */

  req->type = TYPE_OTHER;

  set_status(req, PENDING_COMPLETION);
  set_status(req, PENDING_DATAXFER);
  buffer_lock(req->serial, LOCK_DATAXFER);

  attach_ctio_to_request(ctio, req);

  /* Send response */

  start_next_phase(req);
}


void
do_read (request_t *req)
{
  u_int8_t *buffer;
  int bytes;

  if ((req->cdb[1] & 0x80) && (logic_sock == -1)) {
    send_check_condition(req, SSD_KEY_ILLEGAL_REQUEST, 0x24, 0x00);
    return;
  }

  if (req->cdb[0] == READ_10) {
    struct scsi_rw_10 *rw_10 = (struct scsi_rw_10 *)req->cdb;
    req->offset_blocks = scsi_4btoul(rw_10->addr);
    req->length_blocks = scsi_2btoul(rw_10->length);
  }
  else {
    struct scsi_rw_6 *rw_6 = (struct scsi_rw_6 *)req->cdb;
    req->offset_blocks = scsi_3btoul(rw_6->addr);
    req->length_blocks = rw_6->length;
  }

  if (req->offset_blocks + req->length_blocks > volume_size_blocks) {
    log("Attempt to access past end of volume (%u+%u>%u)\n",
	   req->offset_blocks, req->length_blocks, volume_size_blocks);
    send_check_condition(req, SSD_KEY_ILLEGAL_REQUEST, 0x21, 0x00);
    return;
  }

  if (req->length_blocks == 0) {
    do_nothing(req);
    return;
  }

  req->dataxfer_bytes_remaining = req->length_blocks * 512;
  req->ctio_flags |= CAM_DIR_IN;

  /* Prepare CTIOs that will handle the data transfer */

  req->buffer = allocate_buffer(req->serial, req->dataxfer_bytes_remaining);
  buffer = req->buffer;
  bytes = req->dataxfer_bytes_remaining;

  while (bytes) {
    struct ccb_scsiio *ctio;

    ctio = (struct ccb_scsiio *)get_free_ccb();
    setup_ctio(ctio, req);

    ctio->dxfer_len = min(bytes, DFLTPHYS);
    ctio->data_ptr = buffer;

    bytes -= ctio->dxfer_len;
    buffer += ctio->dxfer_len;

    attach_ctio_to_request(ctio, req);
  }

  /* Activate request */

  if (req->cdb[1] & 0x80) {
    req->type = TYPE_ADMIN_READ;

    set_status(req, PENDING_COMPLETION);
    set_status(req, PENDING_DATAXFER);
    buffer_lock(req->serial, LOCK_DATAXFER);

    assert(logic_sock != -1);
    set_status(req, PENDING_LOGIC);
    buffer_lock(req->serial, LOCK_LOGIC);
  }
  else {
    req->type = TYPE_READ;

    set_status(req, PENDING_COMPLETION);
    set_status(req, PENDING_DATAXFER);
    buffer_lock(req->serial, LOCK_DATAXFER);

    if (logic_sock != -1) {
      set_status(req, PENDING_LOGIC);
      buffer_lock(req->serial, LOCK_LOGIC);
    }

    if (store_sock != -1) {
      set_status(req, PENDING_STORE);
      buffer_lock(req->serial, LOCK_STORE);
    }

    if (clock_enabled == 1) {
      set_status(req, PENDING_CLOCK);
    }
  }

  /* Start request processing */

  start_next_phase(req);
}


void
do_write (request_t *req)
{
  u_int8_t *buffer;
  int bytes;

  if ((req->cdb[1] & 0x80) && (logic_sock == -1)) {
    send_check_condition(req, SSD_KEY_ILLEGAL_REQUEST, 0x24, 0x00);
    return;
  }

  if (req->cdb[0] == WRITE_10) {
    struct scsi_rw_10 *rw_10 = (struct scsi_rw_10 *)req->cdb;
    req->offset_blocks = scsi_4btoul(rw_10->addr);
    req->length_blocks = scsi_2btoul(rw_10->length);
  }
  else {
    struct scsi_rw_6 *rw_6 = (struct scsi_rw_6 *)req->cdb;
    req->offset_blocks = scsi_3btoul(rw_6->addr);
    req->length_blocks = rw_6->length;
  }

  if (req->offset_blocks + req->length_blocks > volume_size_blocks) {
    log("Attempt to access past end of volume (%u+%u>%u)\n",
	   req->offset_blocks, req->length_blocks, volume_size_blocks);
    send_check_condition(req, SSD_KEY_ILLEGAL_REQUEST, 0x21, 0x00);
    return;
  }

  if (req->length_blocks == 0) {
    do_nothing(req);
    return;
  }

  req->dataxfer_bytes_remaining = req->length_blocks * 512;
  req->ctio_flags |= CAM_DIR_OUT;

  /* Prepare CTIOs that will handle the data transfer */

  req->buffer = allocate_buffer(req->serial, req->dataxfer_bytes_remaining);
  buffer = req->buffer;
  bytes = req->dataxfer_bytes_remaining;

  while (bytes) {
    struct ccb_scsiio *ctio;

    ctio = (struct ccb_scsiio *)get_free_ccb();
    setup_ctio(ctio, req);

    ctio->dxfer_len = min(bytes, DFLTPHYS);
    ctio->data_ptr = buffer;

    bytes -= ctio->dxfer_len;
    buffer += ctio->dxfer_len;

    attach_ctio_to_request(ctio, req);
  }

  /* Activate request */

  if (req->cdb[1] & 0x80) {
    req->type = TYPE_ADMIN_WRITE;

    set_status(req, PENDING_COMPLETION);
    set_status(req, PENDING_DATAXFER);
    buffer_lock(req->serial, LOCK_DATAXFER);

    assert(logic_sock != -1);
    set_status(req, PENDING_LOGIC);
    buffer_lock(req->serial, LOCK_LOGIC);
  }
  else {
    req->type = TYPE_WRITE;

    set_status(req, PENDING_COMPLETION);
    set_status(req, PENDING_DATAXFER);
    buffer_lock(req->serial, LOCK_DATAXFER);

    if (logic_sock != -1) {
      set_status(req, PENDING_LOGIC);
      buffer_lock(req->serial, LOCK_LOGIC);
    }

    if (store_sock != -1) {
      set_status(req, PENDING_STORE);
      buffer_lock(req->serial, LOCK_STORE);
    }

    if (clock_enabled == 1) {
      set_status(req, PENDING_CLOCK);
    }
  }

  /* Start request processing */

  start_next_phase(req);
}


void
start_request (struct ccb_accept_tio *atio)
{
  request_t *req;

  req = get_free_request();

  req->ts_arrival = time_stamp();

  req->atio = atio;
  req->ctio_flags = 
    req->atio->ccb_h.flags & (CAM_DIS_DISCONNECT | CAM_TAG_ACTION_VALID);
  req->cdb = (req->ctio_flags & CAM_CDB_POINTER) ?
    req->atio->cdb_io.cdb_ptr : req->atio->cdb_io.cdb_bytes;

  switch (req->cdb[0]) {

  case READ_10:
  case READ_6:
    if (debug) cdb_debug(req->cdb, "READ from %u: ", req->atio->init_id);
    do_read(req);
    break;

  case WRITE_10:
  case WRITE_6:
    if (debug) cdb_debug(req->cdb, "WRITE from %u: ", req->atio->init_id);
    do_write(req);
    break;

  case TEST_UNIT_READY:
    if (debug) cdb_debug(req->cdb, "TEST UNIT READY from %u: ", req->atio->init_id);
    do_nothing(req);
    break;

  case INQUIRY:
    if (debug) cdb_debug(req->cdb, "INQUIRY from %u: ", req->atio->init_id);
    do_inquiry(req);
    break;

  case READ_CAPACITY:
    if (debug) cdb_debug(req->cdb, "READ CAPACITY from %u: ", req->atio->init_id);
    do_read_capacity(req);
    break;

  case REQUEST_SENSE:
  case MODE_SENSE_6:
  case MODE_SELECT_6:
  case SYNCHRONIZE_CACHE:
  case START_STOP_UNIT:
  default:
    log("unhandled CDB: %#x\n", req->cdb[0]);
    send_check_condition(req, SSD_KEY_ILLEGAL_REQUEST, 0x24, 0x00);
    /* GRONK */
  }

  calibration_time_stamp_arrive(req);
}


/************************************************************************
 * Event processing
 ************************************************************************/

void
try_to_complete_request (request_t *req)
{
  switch (req->type) {
  case TYPE_INTERPOSE:
    assert(!req->status);
    req->ts_completion = time_stamp();
    print_request_times(req);
    release_request(req);
    return;

  case TYPE_READ:
  case TYPE_ADMIN_READ:
  case TYPE_OTHER:
    if (req->status == PENDING_COMPLETION) {
      clear_status(req, PENDING_COMPLETION);
      set_status(req, SENT_COMPLETION);
      req->ts_completion = time_stamp();
      calibration_time_stamp_complete(req);
      send_request_complete(req);
    }
    return;

  case TYPE_WRITE:
  case TYPE_ADMIN_WRITE:
    if ((req->status & ~(PENDING_STORE & STARTED_STORE)) 
	== PENDING_COMPLETION) {
      clear_status(req, PENDING_COMPLETION);
      set_status(req, SENT_COMPLETION);
      req->ts_completion = time_stamp();
      calibration_time_stamp_complete(req);
      send_request_complete(req);
    }
    return;

  default:
    log("Bad req->type %x\n", req->type);
    exit(1);
  }
}


void
handle_ctio_complete (struct ccb_scsiio *ctio)
{
  request_t *req;
  box_t *this;
  // box_t *next;

  req = get_request_by_ctio(ctio);

  if (debug) log("request %d, ctio %p\n", req->serial, ctio);

  this = TAILQ_FIRST(&req->ctios_pending);
  TAILQ_REMOVE(&req->ctios_pending, this, tqe);
  assert(this->u.ctio == ctio);

  /* If we've sent the request completion, there is nothing more to do
   * than free up the resources from this request.
   */

  if (req->status == SENT_COMPLETION) {
    print_request_times(req);
    // calibration_time_stamp_complete(req);
    release_request(req);
    goto end;
  }

  /* If there are still bytes to transfer to or from the kernel, then
   * all we do is start the next CTIO and exit the funciton.
   */

  req->dataxfer_bytes_remaining -= ctio->dxfer_len;
  if (req->dataxfer_bytes_remaining) {
    goto end;
  }

  req->ts_dataxfer_end = time_stamp();

  clear_status(req, STARTED_DATAXFER);
  if (req->buffer) buffer_unlock(req->serial, LOCK_DATAXFER);

  start_next_phase(req);

 end:

  release_ccb((union ccb *)this->u.ctio);
  release_box(this);
}


void
handle_ccb (void)
{
  union ccb *ccb_array[MAX_INITIATORS], *ccb;
  int ccb_count, i;

  ccb_count = read(targ_fd, ccb_array, sizeof(ccb_array));
  if (ccb_count <= 0) {
    log("read ccb ptrs: %s\n", strerror(errno));
    assert(0);
  }
  ccb_count /= sizeof(union ccb *);
  if (ccb_count < 1) {
    log("%s\n","truncated read ccb ptr");
    assert(0);
  }

  for (i=0; i<ccb_count; i++) {
    ccb = ccb_array[i];

    switch (ccb->ccb_h.func_code) {

    case XPT_ACCEPT_TARGET_IO:
      start_request(&ccb->atio);
      break;

    case XPT_CONT_TARGET_IO:
      handle_ctio_complete(&ccb->ctio);
      break;

    default:
      log("Unhandled ccb type %#x\n", ccb->ccb_h.func_code);
    }
  }
}


void
handle_logic_timeout (int serial)
{
  request_t *req;

  log("\n\
------------------------------------------------------------------------\n\
Logic timeout for request %d; closing logic socket\n\
------------------------------------------------------------------------\n",
      serial);

  /* Close the logic socket if it's not already closed */

  if (logic_sock != -1) {
    if (close(logic_sock) == -1) log ("close logic: %s\n", strerror(errno));
    logic_sock = -1;
  }

  /* Finish this request as though we'd received an IPC_TYPE_REQUEST_OK */
  
  req = get_request_by_serial(serial);
  if (!req) {
    if (!req) {
      log("%s\n", "The logic timer completed for a request but it doesn't exist.  This should never happen!  The timer must not be getting cleared when the logic finishes.");
      exit(1);
    }
  }

  clear_status(req, STARTED_LOGIC);
  buffer_unlock(req->serial, LOCK_LOGIC);

  start_next_phase(req);
}


void
handle_logic (ipc_message_t *msg_in)
{
  request_t *req;

  if (debug) log("typ %d, ser %d, off %lld, len %d\n", msg_in->type, msg_in->serial, msg_in->request_offset_blocks, msg_in->request_length_blocks);

  switch (msg_in->type) {

  case IPC_TYPE_REQUEST_OK:
    req = get_request_by_serial(msg_in->serial);
    if (!req || !(req->status & STARTED_LOGIC)) {
      /* If the request doesn't exist anymore, or if the "STARTED_LOGIC"
       * flag has been cleared (which happens when the timer has already
       * expired) then do nothing. */
      log("The logic module completed request %d, but it had already timed out.\n", msg_in->serial);
      break;
    }

    req->ts_logic_recv = time_stamp();
    req->ts_logic_begin = msg_in->ts_begin;
    req->ts_logic_end = msg_in->ts_end;

    clear_status(req, STARTED_LOGIC);
    buffer_unlock(req->serial, LOCK_LOGIC);

    start_next_phase(req);

    if (req->type != TYPE_INTERPOSE) {
      clear_logic_timeout(req->serial);
    }
    break;

  case IPC_TYPE_INTERPOSE_READ:
    req = get_free_request();

    req->ts_arrival = time_stamp();

    req->offset_blocks = msg_in->request_offset_blocks;
    req->length_blocks = msg_in->request_length_blocks;
    req->buffer = allocate_buffer(req->serial, req->length_blocks * 512);

    /* Activate request */

    req->type = TYPE_INTERPOSE;

    /* GRONK: make this a priority read? */
    if (store_sock != -1) {
      set_status(req, PENDING_STORE);
      buffer_lock(req->serial, LOCK_STORE);
    }

    set_status(req, PENDING_LOGIC);
    buffer_lock(req->serial, LOCK_LOGIC);

    /* Start request processing */

    start_next_phase(req);
    break;

  default:
    log("unknown type: %d\n", msg_in->type);
    assert(0);
  }
}


void
handle_store (ipc_message_t *msg_in)
{
  request_t *req;

  if (debug) log("typ %d, ser %d, off %lld, len %d\n", msg_in->type, msg_in->serial, msg_in->request_offset_blocks, msg_in->request_length_blocks);

  switch (msg_in->type) {

  case IPC_TYPE_REQUEST_OK:
    req = get_request_by_serial(msg_in->serial);
    if (!req) {
      log("%s\n", "The store completed a request but it doesn't exist.  This should never happen!  Did the store try to complete this request twice?");
      exit(1);
    }

    req->ts_store_recv = time_stamp();
    req->ts_store_begin = msg_in->ts_begin;
    req->ts_store_end = msg_in->ts_end;

    clear_status(req, STARTED_STORE);
    buffer_unlock(req->serial, LOCK_STORE);

    start_next_phase(req);
    break;

  case IPC_TYPE_PARAMETERS:
    volume_size_blocks = msg_in->request_length_blocks;
    if (debug) log("volume_size_blocks=%d\n", volume_size_blocks);
    break;

  default:
    log("unknown type: %d\n", msg_in->type);
    assert(0);
  }
}


void
handle_clock (request_t *req)
{
  if (debug) log("type %#x: serial %d: %#x\n", req->type, req->serial, req->status);

  clear_status(req, STARTED_CLOCK);

  if ((req->type == TYPE_READ) && (req->status != PENDING_COMPLETION)) {
    log("%d cannot complete on time: status %#x \n", req->serial, req->status);
  }

  try_to_complete_request(req);
}


void
handle_ipc (int sock)
{
  int rc;
  ipc_message_t msg_in;

  memset(&msg_in, 0, sizeof(ipc_message_t));
  rc = recv_ipc(sock, &msg_in, sizeof(ipc_message_t));
  if (rc != sizeof(ipc_message_t)) {
    log("bad length %d\n", rc);
    return;
  }

  if (msg_in.magic != IPC_MAGIC) {
    log("bad magic %#x\n", msg_in.magic);
    return;
  }

  if (debug) log("source %d, type %d, serial %d\n", msg_in.source, msg_in.type, msg_in.serial);

  switch (msg_in.source) {

  case IPC_SOURCE_LOGIC:
    assert(sock == logic_sock);
    handle_logic(&msg_in);
    break;

  case IPC_SOURCE_STORE:
    assert(sock == store_sock);
    handle_store(&msg_in);
    break;

  default:
    log("unknown socket %d\n", sock);
    assert(0);
  }
}


void
handle_eof (int sock)
{
  const int max_events=1;
  struct kevent events[max_events];
  int n=0;
  struct request *req;

  EV_SET(&events[n++], sock, EVFILT_READ, EV_DELETE, 0,0,0);
  assert(n<=max_events);
  if (kevent(kq_fd, events, n, NULL, 0, NULL) == -1) {
    log("kevent: %s\n", strerror(errno));
    assert(0);
  }

  if (sock == store_sock) {
    log("%s disconnected unexpectedly\n", "store");
    if (close(store_sock) == -1) log("close store: %s\n", strerror(errno));
    store_sock = -1;
    volume_size_blocks = 0;
    /* loop through all requests that are pending at the module.
     * pretend the logic has completed (including unlocking the buffer),
     * and start the next event. */
    TAILQ_FOREACH(req, &active_reqs, tqe) {
      if (req->status & PENDING_STORE) {
	clear_status(req, PENDING_STORE);
	buffer_unlock(req->serial, LOCK_STORE);
      }
      if (req->status & STARTED_STORE) {
	clear_status(req, STARTED_STORE);
	buffer_unlock(req->serial, LOCK_STORE);
	start_next_phase(req);
      }
    }
  }
  else if (sock == logic_sock) {
    log("%s disconnected unexpectedly\n", "logic");
    if (close(logic_sock) == -1) log("close logic: %s\n", strerror(errno));
    logic_sock = -1;
    /* loop through all requests that are pending at the module.
     * pretend the logic has completed (including unlocking the buffer),
     * and start the next event. */
    TAILQ_FOREACH(req, &active_reqs, tqe) {
      if (req->status & PENDING_LOGIC) {
	clear_status(req, PENDING_LOGIC);
	buffer_unlock(req->serial, LOCK_LOGIC);
      }
      if (req->status & STARTED_LOGIC) {
	clear_status(req, STARTED_LOGIC);
	buffer_unlock(req->serial, LOCK_LOGIC);
	start_next_phase(req);
      }
    }
  }
  else {
    log("unknown socket %d\n", sock);
  }
}


/************************************************************************
 * Timing manager functions
 ************************************************************************/

#ifdef USE_DISKSIM

int clock_times_fd = -1;

void
clock_open_times_file (void)
{
  const int filename_maxlen=80;
  char filename[filename_maxlen];

  if (!calibration_prefix) return;

  if ((strlen(calibration_prefix) + 4) > filename_maxlen) {
    log("please increase filename_maxlen (%d)\n", filename_maxlen);
    assert(0);
  }

  snprintf(filename, filename_maxlen,
	   "%s.sim",
	   calibration_prefix);

  clock_times_fd = open(filename, O_CREAT|O_WRONLY, 0644);
  if (clock_times_fd == -1) {
    log("unable to open %s: %s\n", filename, strerror(errno));
    assert(0);
  }
}


void
clock_close_times_file (void)
{
  if (clock_times_fd == -1) return;

  if (close(clock_times_fd) == -1) {
    log("close: %s\n", strerror(errno));
  }
  clock_times_fd = -1;
}


void
clock_times_file_write (double time, struct disksim_request *dr)
{
  const int buflen = 160;
  unsigned char buf[buflen];
  int bytes;

  if (clock_times_fd == -1) return;

  bytes = snprintf(buf, buflen,
		   "%ld\t%d\t%x\t%x\t%f\t%d\n",
		   dr->blkno,
		   dr->bytecount / 512,
		   (dr->flags & DISKSIM_READ) ? 0 : 1,  /* 0 for read */
		   MEMULATOR_EVENT_TARGET_SIMULATOR_TIME,
		   time - dr->start + skew_ms,
		   ((request_t *)dr->reqctx)->serial);

  write(clock_times_fd, buf, bytes);
}


u_int64_t clock_timestamp_calibration = 1993868931LL;  
                                         /* four-day calibrated average */
u_int64_t clock_timestamp_simstart = 0;  /* RDTSC when simulation started */

double clock_prev_event_simtime = 0.0;  /* Time when prev event fired */
double clock_next_event_simtime = 0.0;  /* Time when next event fires */
disksim_interface_callback_t clock_next_event_callback = NULL;
                                      /* Callback when next event fires */

struct disksim_interface *clock_disksim_interface = NULL;
                                      /* Context for simulator functions */

int clock_requests_pending = 0;  /* Count of # of requests in simulator */

double clock_next_completion_simtime = 0.0;
TAILQ_HEAD(, request) clock_queue_complete = 
                        TAILQ_HEAD_INITIALIZER(clock_queue_complete);


inline double
clock_get_curtime (void)
{
  u_int64_t ts_now;
  u_int64_t ts_elapsed;

  ts_now = time_stamp();
  ts_elapsed = ts_now - clock_timestamp_simstart;

  /* convert to msec with the multiply-by-1000 */
  return ((double)(ts_elapsed) / (double)clock_timestamp_calibration) * 1000.0;
}


/* callback (disksim -> client) to register a callback (client ->
 * disksim); when simulated time reaches t, client code should invoke
 * fn. */
void
clock_schedule_callback (disksim_interface_callback_t fn,
		       double time_msec,
		       void *context)
{
  clock_next_event_simtime = time_msec;
  clock_next_event_callback = fn;
}


/* cancel the pending callback set for t */
void
clock_deschedule_callback (double time_msec,
			 void *context)
{
  clock_next_event_simtime = 0.0;
  clock_next_event_callback = NULL;
}


void
clock_request_arrive (request_t *req)
{
  struct disksim_request *dr;
  double interface_start;

  if (debug) log("request %d arrival in clock module\n", req->serial);

  dr = xcalloc(1, sizeof(struct disksim_request));
  switch (req->type) {
  default:
    log("WARNING: Unknown request type %d; reverting to read!\n", req->type);
    /* fallthrough */
  case TYPE_READ:
  case TYPE_INITIALIZER:
    dr->flags = DISKSIM_READ;
    break;
  case TYPE_WRITE:
    dr->flags = DISKSIM_WRITE;
    break;
  }
  dr->devno          = 0;
  dr->blkno          = req->offset_blocks;
  dr->bytecount      = req->length_blocks * 512;
  dr->start          = clock_get_curtime();
  dr->batchno        = 0;
  dr->batch_complete = 0;
  dr->reqctx         = (void *)req;

  clock_requests_pending++;

  if (clock_prev_event_simtime > dr->start) {
    interface_start = clock_prev_event_simtime;
  } else {
    interface_start = dr->start;
  }

  disksim_interface_request_arrive(clock_disksim_interface, interface_start, dr);
}


int
clock_attempt_next_event (void)
{
  double curtime;  /* Current time in milliseconds */
  int loop_again;

 loop:
  curtime = clock_get_curtime();
  loop_again = 0;

  /* Execute the simulation under the current-time modification
   * technique, but don't advance the clock past the next completion
   * time if not necessary. */

  if (clock_next_completion_simtime == 0.0
      && clock_next_event_simtime > 0.0
      && clock_next_event_simtime <= (curtime + skew_ms + lookahead_ms)) {
    clock_prev_event_simtime = clock_next_event_simtime;
    clock_next_event_simtime = 0.0;
    clock_next_event_callback(clock_disksim_interface, clock_prev_event_simtime, 0);
    loop_again = 1;  /* OK to loop to see if another event is ready */
  }

  /* If there are any requests that are to complete within the next
   * (lookahead_ms) milliseconds, we will go into a tight loop until it's
   * time for that request to complete. */

  if (clock_next_completion_simtime > 0.0) {
    if (clock_next_completion_simtime <= (curtime + skew_ms)) {
      request_t *r = TAILQ_FIRST(&clock_queue_complete);
      TAILQ_REMOVE(&clock_queue_complete, r, clock_complete_tqe);
      handle_clock(r);

      if (TAILQ_EMPTY(&clock_queue_complete)) {
	clock_next_completion_simtime = 0.0;
      } else {
	request_t *next = TAILQ_FIRST(&clock_queue_complete);
	clock_next_completion_simtime = next->clock_completion_simtime;
      }

      return (1);  /* Return now, don't allow sched_yield() after return */
    }
    else if (clock_next_completion_simtime <= 
	     (curtime + skew_ms + busywait_ms)) {
      loop_again = 1;  /* Busy wait if the completion is imminent */
    }
  }

  if (loop_again) {
    goto loop;
  }

  return (0);  /* OK for sched_yield() after return */
}


void
clock_request_complete (double time,
			struct disksim_request *dr,
			void *context)
{
  request_t *r;

  clock_requests_pending--;

  r = (request_t *)dr->reqctx;
  if (r->type == TYPE_INITIALIZER) {
    release_request(r);
    return;
  }

  r->clock_completion_simtime = time;
  /* Insert request onto tail-q of completed (timing) requests */
  TAILQ_INSERT_TAIL(&clock_queue_complete, r, clock_complete_tqe);
  if (clock_next_completion_simtime == 0.0) {
    clock_next_completion_simtime = time;
  }

  clock_times_file_write(time, dr);
}


void
clock_start_simulation (void)
{
  request_t *r;

  clock_open_times_file();

  clock_disksim_interface = 
    disksim_interface_initialize("parfile",
				 "outfile",
				 clock_request_complete,
				 clock_schedule_callback,
				 clock_deschedule_callback,
				 0);
  clock_timestamp_simstart = time_stamp();

  /* Send zero-length request to disksim to get things warmed up. */
  r = get_free_request();
  r->offset_blocks = 0;
  r->length_blocks = 1;
  r->type = TYPE_INITIALIZER;

  clock_request_arrive(r);
}


void
clock_stop_simulation (void)
{
  double time = clock_get_curtime();
  if (time > clock_next_event_simtime) {
    time = clock_next_event_simtime;
  }

  disksim_interface_shutdown(clock_disksim_interface, time);

  clock_close_times_file();
}


#else  /* ifdef USE_DISKSIM */

#define CLOCK_KILL {	\
  log("%s\n",		\
      "FATAL: this should never be called if USE_DISKSIM is not defined!"); \
  exit(1);		\
}

int clock_attempt_next_event (void) { CLOCK_KILL; }
void clock_request_arrive (request_t *req) { CLOCK_KILL; }
void clock_start_simulation (void) { CLOCK_KILL; }
void clock_stop_simulation (void) { CLOCK_KILL; }

#endif  /* ifdef USE_DISKSIM */


/************************************************************************
 * Main loop
 ************************************************************************/

int
main_loop (void)
{
  const int max_events=MAX_INITIATORS+5;
  struct kevent events[max_events];
  int n=0;
  int signal_received=0;

  /* Register kqueue for event notification */
  if ((kq_fd = kqueue()) == -1) {
    log("kqueue: %s\n", strerror(errno));
    assert(0);
  }

  signal(SIGHUP, SIG_IGN);
  signal(SIGINT, SIG_IGN);
  signal(SIGTERM, SIG_IGN);
  signal(SIGPROF, SIG_IGN);
  signal(SIGALRM, SIG_IGN);
  signal(SIGSTOP, SIG_IGN);
  signal(SIGTSTP, SIG_IGN);
  signal(SIGUSR1, SIG_IGN);
  signal(SIGPIPE, SIG_IGN);

  /* Set up some default events */
  EV_SET(&events[n++], SIGHUP,  EVFILT_SIGNAL, EV_ADD, 0,0,0);
  EV_SET(&events[n++], SIGINT,  EVFILT_SIGNAL, EV_ADD, 0,0,0);
  EV_SET(&events[n++], SIGTERM, EVFILT_SIGNAL, EV_ADD, 0,0,0);
  EV_SET(&events[n++], SIGUSR1, EVFILT_SIGNAL, EV_ADD, 0,0,0);
  if (targ_fd != -1) {
    EV_SET(&events[n++], targ_fd, EVFILT_READ, EV_ADD, 0,0,0);
  }
  if (listen_sock != -1) {
    EV_SET(&events[n++], listen_sock, EVFILT_READ, EV_ADD, 0,0,0);
  }

  assert(n<=max_events);
  if (kevent(kq_fd, events, n, NULL, 0, NULL) == -1) {
    log("kevent 1: %s\n", strerror(errno));
    assert(0);
  }

  while (!signal_received) {
    int num_events, i;
    struct timespec ts = {0, 0};

    /* Check for the next signal or read ready */
    /* GRONK: example code uses clever value for timeout */

    num_events = kevent(kq_fd, NULL, 0, events, max_events, &ts);
    if (num_events == -1) {
      if (errno == EINTR) continue;
      log("kevent 2: %s\n", strerror(errno));
      assert(0);
    }
    if (num_events > max_events) {
      log("kevent 3: %s\n", strerror(errno));
      assert(0);
    }

    /* Process all received events */

    for (i=0; i<num_events; i++) {
      if ((events[i].flags & EV_ERROR) != 0) {
	log("kevent 4: %s\n", strerror(errno));
	assert(0);
      }

      switch (events[i].filter) {

      case EVFILT_READ:
	if (events[i].ident == targ_fd) {
	  handle_ccb();
	}
	else if (events[i].ident == store_sock
		 || events[i].ident == logic_sock) {
	  if (!(events[i].flags & EV_EOF)) {
	    handle_ipc(events[i].ident);
	  }
	  else {  /* Other side disconnected */
	    handle_eof(events[i].ident);
	  }
	}
	else if (events[i].ident == listen_sock) {
	  handle_new_ipc_connection();
	}
	else {
	  log("EVFILT_READ: unknown ident %#x\n", events[i].ident);
	}
	break;

      case EVFILT_SIGNAL:
	if (debug) log("signal received (%#x)\n", events[i].ident);
	if ((events[i].ident == SIGHUP)
	    || (events[i].ident == SIGINT)
	    || (events[i].ident == SIGKILL)) {
	  signal_received=1;
	}
	else if (events[i].ident == SIGUSR1) {
	  if (debug) log("cycling times file (%d)\n", times_number);
	  close_times_file();
	  open_times_file();
	}
	break;

      case EVFILT_TIMER:
	if (debug) log("timer fired (%#x)\n", events[i].ident);
	handle_logic_timeout(events[i].ident);
	break;

      default:
	log("unknown event %#x\n", events[i].filter);
	break;
      }
    }

    if (clock_enabled == 1) {
      num_events += clock_attempt_next_event();
    }

    if (1 || num_events == 0) {
      sched_yield();
    }
  }

  return (0);  /* GRONK */
}


/************************************************************************
 * Functions for interaction with device driver
 ************************************************************************/

void
enable_target_mode_driver (void)
{
  assert(targ_fd == -1);

  if (ioc_enlun.path_id == CAM_BUS_WILDCARD
      || ioc_enlun.target_id == CAM_TARGET_WILDCARD
      || ioc_enlun.lun_id == CAM_LUN_WILDCARD) {
    log("Invalid target path specified: %d:%d:%d\n",
	   ioc_enlun.path_id, ioc_enlun.target_id, ioc_enlun.lun_id);
    assert(0);
  }

  /* Open communications path with device driver */
  {
    int unit=0;
    char targname[80];

    do {
      snprintf(targname, sizeof(targname), "/dev/targ%d", unit++);
      targ_fd = open(targname, O_RDWR);
    } while (targ_fd == -1 && errno == EBUSY);

    if (targ_fd == -1) {
      log("Tried to open %d devices, none available", unit);
      assert(0);
    }
  }

  /* Enable listening on the specified LUN */

  if (ioctl(targ_fd, TARGIOCENABLE, &ioc_enlun) == -1) {
    log("ioctl TARGIOCENABLE: %s\n", strerror(errno));
    assert(0);
  }

  /* Detect capabilities supported by the SCSI interface module (SIM) */
  {
    struct ccb_pathinq cpi;
    cam_status status;

    memset(&cpi, 0, sizeof(cpi));
    cpi.ccb_h.func_code = XPT_PATH_INQ;
    send_ccb((union ccb *)&cpi);
    status = cpi.ccb_h.status & CAM_STATUS_MASK;

    if (status != CAM_REQ_CMP) {
      log("CAM path inquiry failed, status %#x\n", status);
      assert(0);
    }

    /* Can only enable on controllers that support target mode */
    if ((cpi.target_sprt & PIT_PROCESSOR) == 0) {
      log("%s\n","HBA does not support target mode");
      assert(0);
    }

    sim_flags = cpi.hba_inquiry;
  }

  /* Let's see what we can do to improve the Adaptec transfer time. */
  req_flags |= SID_Sync;
  req_flags |= SID_WBus16;

  /* Queue ATIOs at device driver.  ATIOs (Accept Target
   * I/Os) are used to notify us when a new target-mode request arrives.
   * When we are completely finished with the request, we send the ATIO
   * back to the driver so it can reuse it.
   */
  {
    int i;

    for (i=0; i<MAX_INITIATORS; i++) {
      struct ccb_accept_tio *atio;

      atio = (struct ccb_accept_tio *)get_free_ccb();
      atio->ccb_h.func_code = XPT_ACCEPT_TARGET_IO;
      send_ccb((union ccb *)atio);

      /* GRONK: Example code also sends INOTs here. */
    }
  }
}


void
disable_target_mode_driver (void)
{
  assert(targ_fd != -1);

  if (ioctl(targ_fd, TARGIOCDISABLE, NULL) == -1) {
    log("ioctl TARGIOCDISABLE: %s\n", strerror(errno));
  }
  if (close(targ_fd) == -1) {
#if 0
    log("close: %s\n", strerror(errno));
#endif
  }

  targ_fd = -1;
}


/************************************************************************
 * Main routines
 ************************************************************************/

void
process_args (int argc, char **argv)
{
  char *program = argv[0];
  int ch;

  while ((ch = getopt(argc, argv, "b:cdf:k:l:m:r:t:x:y:z:")) != -1) {
    switch (ch) {
    case 'b':  /* backing store filename */
      store_filename=optarg;
      break;
    case 'c':
#ifndef USE_DISKSIM
      printf("\
------------------------------------------------------------------------\n\
You must define USE_DISKSIM in the source code to use the -c option\n\
------------------------------------------------------------------------\n");
      exit(1);
#endif  /* ifndef USE_DISKSIM */
      log("%s\n", "Clock enabled.");
      clock_enabled = 1;
      break;
    case 'd':
      debug = 1;
      break;
    case 'f':
      calibration_prefix = optarg;
      break;
    case 'k':
#ifndef USE_DISKSIM
      printf("\
------------------------------------------------------------------------\n\
You must define USE_DISKSIM in the source code to use the -k option\n\
------------------------------------------------------------------------\n");
      exit(1);
#endif  /* ifndef USE_DISKSIM */
      skew_ms = atof(optarg) / 1000.0;
      log("Skew set to %f ms\n", skew_ms);
      break;
    case 'l':
#ifndef USE_DISKSIM
      printf("\
------------------------------------------------------------------------\n\
You must define USE_DISKSIM in the source code to use the -l option\n\
------------------------------------------------------------------------\n");
      exit(1);
#endif  /* ifndef USE_DISKSIM */
      lookahead_ms = atof(optarg) / 1000.0;
      log("Lookahead set to %f ms\n", lookahead_ms);
      break;
    case 'm':
#ifndef USE_DISKSIM
      printf("\
------------------------------------------------------------------------\n\
You must define USE_DISKSIM in the source code to use the -m option\n\
------------------------------------------------------------------------\n");
      exit(1);
#endif  /* ifndef USE_DISKSIM */
      busywait_ms = atof(optarg) / 1000.0;
      log("Busy-wait set to %f ms\n", busywait_ms);
      break;
    case 'r':
      trace_dirname = optarg;
      break;
    case 't':
      times_filename=optarg;
      break;
    case 'x':
      memory_preload_file = optarg;
      break;
    case 'y':
      store_memory_size = optarg;
      break;
    case 'z':
      volume_size_blocks = atoi(optarg);
      break;
    default:
      goto usage;
    }
  }
  argc -= optind;
  argv += optind;

  if (argc != 1) {
    printf("argc==%d\n", argc);
  usage:
    printf("\
usage: \n\
  %s [-b file] [-c] [-d] [-f prefix] \n\
     [-k time] [-l time] [-r dir] [-t file] [-z size] \n\
     bus:target:lun \n\
\n\
  where: \n\
  -b: start a store procss, using <file> for the backing store \n\
  -c: use disksim for simulated timings (must define USE_DISKSIM in code) \n\
  -d: verbose debugging output \n\
  -f: this is the prefix for request timestamp output files \n\
  -k: use <time> microseconds for skew in simulated timings (must use -c) \n\
  -l: use <time> microseconds for current-time lookahead (must use -c) \n\
  -r: keep a trace (plus contents) of all block requests \n\
  -t: print times output to filename <file.001> \n\
  -y: start a store process, using <size> as the local memory size \n\
  -z: force capacity to be reported as <size> (doesn't work with -b)\n\
\n\
  standard usage for IDD will look something like: \n\
     %s -b /dev/da2 1:96:0 \n\
", program, program);
    exit(1);
  }

  memset(&ioc_enlun, 0, sizeof(ioc_enlun));
  sscanf(argv[0], "%u:%u:%u",
	 &ioc_enlun.path_id, &ioc_enlun.target_id, &ioc_enlun.lun_id);
  /* We don't support any vendor-specific commands */
  ioc_enlun.grp6_len = 0;
  ioc_enlun.grp7_len = 0;
}


int
main (int argc, char **argv)
{
  int ret=0;

  process_args(argc, argv);

  open_calibration_times_file();
  open_times_file();

  enable_target_mode_driver();

  enable_ipc();

  start_store();

  if (clock_enabled) clock_start_simulation();

  ret = main_loop();

  if (clock_enabled) clock_stop_simulation();

  stop_store();

  disable_ipc();

  disable_target_mode_driver();

  close_times_file();
  close_calibration_times_file();

  return (ret);
}
