/*
 * The "Memulator" timing-accurate storage emulation environment
 * Authors: John Linwood Griffin, John Bucy, Greg Ganger
 *
 * Copyright (c) of Carnegie Mellon University, 2001-2004.
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 *
 */

#define MEMULATOR_EVENT_TARGET_SIMULATOR_TIME    (0x40)

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <aio.h>
#include <errno.h>
#include <fcntl.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/event.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/time.h>
#include <sys/queue.h>

#include "memulator.h"
#define USE_SIMULATION
#ifdef USE_SIMULATION
#include "disksim/src/disksim_interface.h"
#endif


static int debug=0;
static int kq_fd=-1;  /* Descriptor for kernel event queue */
static int memulator_sock=-1;  /* Socket for IPC with memulator */
static int received_termination=0;  /* When memulator tells us to shut down */

static char *calibration_prefix = NULL;
static int calibration_fd = -1;

/************************************************************************
 * Support functions
 ************************************************************************/

#define log(a,b...) \
   do { fprintf(stdout, "%s: %s: " a, "clock", __func__, b); \
        fflush(stdout); } while (0)


static inline void
time_stamp (u_int64_t *d)
{
  asm("rdtsc" : "=&A" (*d));
}


void *
xcalloc (size_t size)
{
  register void *value = calloc(1, size);
  if (value == 0) {
    log("unable to allocate %d bytes, exiting", size);
    assert(0);
  }
  return value;
}


/************************************************************************
 * Timings output
 ************************************************************************/

void
open_calibration_times_file (void)
{
  const int filename_maxlen=80;
  char filename[filename_maxlen];

  if (!calibration_prefix) return;

  if ((strlen(calibration_prefix) + 4) > filename_maxlen) {
    log("please increase filename_maxlen (%d)\n", filename_maxlen);
    assert(0);
  }

  snprintf(filename, filename_maxlen,
	   "%s.sim",
	   calibration_prefix);

  calibration_fd = open(filename, O_CREAT|O_WRONLY, 0644);
  if (calibration_fd == -1) {
    log("unable to open %s: %s\n", filename, strerror(errno));
    assert(0);
  }
}


void
close_calibration_times_file (void)
{
  if (calibration_fd == -1) return;

  if (close(calibration_fd) == -1) {
    log("close: %s\n", strerror(errno));
  }
  calibration_fd = -1;
}


void
calibration_time_stamp (double time, struct disksim_request *r)
{
  struct event {
    u_int32_t offset;
    u_int16_t length;
    u_int8_t  opcode;
    u_int8_t  code;
    double    timestamp;
  } ev;  /* 128 bits */

  if (calibration_fd == -1) return;

  ev.offset    = r->blkno;
  ev.length    = r->bytecount / 512;
  ev.opcode    = (r->flags & DISKSIM_WRITE) ? 1 : 0;  /* 0 for read */
  ev.code      = MEMULATOR_EVENT_TARGET_SIMULATOR_TIME;
  ev.timestamp = time - r->start;

  write(calibration_fd, &ev, sizeof(struct event));
}


/************************************************************************
 * Interprocess communication
 ************************************************************************/

static int
send_ipc (int sock, void *buf, size_t len)
{
  u_int8_t *msg;
  int bytes = 0;	/* Total bytes sent */
  int rc;

  if (debug) log("%s\n","send_ipc");

  msg = buf;

  while (bytes < len) {
    if ((rc = send(sock, msg + bytes, len - bytes, 0)) == -1) {
      log("send: %s", strerror(errno));
      return (-1);
    }
    bytes += rc;
  }

  if (bytes != len) {
    log("bad transfer (%d != %d)\n", bytes, len);
  }

  return (bytes);
}


static int
recv_ipc (int sock, void *buf, size_t len)
{
  u_int8_t *msg;
  int bytes = 0;	/* Total bytes received */
  int rc;

  if (debug) log("%s\n","recv_ipc");

  msg = buf;

  while (bytes < len) {
    if ((rc = recv(sock, buf + bytes, len - bytes, 0)) == -1) {
      log("recv: %s", strerror(errno));
      return (-1);
    }
    bytes += rc;
  }

  if (bytes != len) {
    log("bad transfer (%d != %d)\n", bytes, len);
  }

  return (bytes);
}


static void
memulator_attach (void)
{
  struct sockaddr_un sun;
  size_t size;
  ipc_message_t msg_out;

  assert(memulator_sock==-1);

  /* socket */

  if ((memulator_sock = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) {
    log("socket: %s", strerror(errno));
    assert(0);
  }

  /* connect */

  memset(&sun, 0, sizeof(struct sockaddr_un));
  sun.sun_family = AF_UNIX;
  strncpy(sun.sun_path, MEMULATOR_SOCKET_FILENAME, 104);
  size = SUN_LEN(&sun);
  if (connect(memulator_sock, (struct sockaddr *)&sun, size) == -1) {
    log("connect: %s\n", strerror(errno));
    assert(0);
  }

  /* Send connect message */

  memset(&msg_out, 0, sizeof(ipc_message_t));
  msg_out.magic  = IPC_MAGIC;
  msg_out.source = IPC_SOURCE_CLOCK;
  msg_out.type   = IPC_TYPE_CONNECT;
  send_ipc(memulator_sock, &msg_out, sizeof(ipc_message_t));
}


/************************************************************************
 * Simulation support
 ************************************************************************/

static struct disksim_interface *disksim_interface;
u_int64_t ts_simulation_start;
static int num_pending_requests=0;

disksim_interface_callback_t disksim_callback_fn = NULL;


static inline double
get_simtime_msec (void)
{
  /* GRONK -- assumes nc94.
   * 
   * nc93: 1993907191
   * nc94: 1993868287
   * nc95: 1993925703
   * nc96: 1993905323
   */

  u_int64_t ts_now;
  register double simtime_msec;

  time_stamp(&ts_now);

  simtime_msec = (ts_now - ts_simulation_start) / (1883868.2870);

  if (debug) log("simtime_msec = %f\n", simtime_msec);
  return simtime_msec;
}


void
handle_request_arrive (ipc_message_t *msg_in)
{
  struct disksim_request *r;

  if (debug) log("request %d arrival\n", msg_in->serial);

  r = xcalloc(sizeof(struct disksim_request));
  if (msg_in->type == IPC_TYPE_READ) {
    r->flags |= DISKSIM_READ;
  } else if (msg_in->type == IPC_TYPE_WRITE) {
    r->flags |= DISKSIM_WRITE;
  } else {
    assert(0);
  }
  r->devno          = 0;
  r->blkno          = msg_in->request_offset_blocks;
  r->bytecount      = msg_in->request_length_blocks * 512;
  r->start          = get_simtime_msec();
  r->batchno        = 0;
  r->batch_complete = 0;
  r->reqctx         = (void *)msg_in->serial;

  num_pending_requests++;

  disksim_interface_request_arrive(disksim_interface, r->start, r);
}


/* callback when requests complete */
void
handle_request_complete (double time,
			 struct disksim_request *r,
			 void *context)
{
  ipc_message_t msg_out;

  if (debug) log("request %d complete\n", (int)r->reqctx);

  memset(&msg_out, 0, sizeof(ipc_message_t));
  msg_out.magic  = IPC_MAGIC;
  msg_out.source = IPC_SOURCE_CLOCK;
  msg_out.type   = IPC_TYPE_REQUEST_OK;
  msg_out.serial = (int)r->reqctx;
  send_ipc(memulator_sock, &msg_out, sizeof(ipc_message_t));

  calibration_time_stamp(time, r);

  num_pending_requests--;
  free(r);  r = NULL;
}


/* callback (disksim -> client) to register a callback (client ->
 * disksim); when simulated time reaches t, client code should invoke
 * fn. */
void
schedule_callback (disksim_interface_callback_t fn,
		   double time,
		   void *context)
{
  double now;
  double difference;
  int sec;
  int usec;
  struct itimerval value;

  if (debug) log("scheduling callback %f\n", time);

  assert(disksim_callback_fn == NULL);

  now = get_simtime_msec();
  difference = time - now;

  disksim_callback_fn = fn;

  if (difference > 0.001) {
    sec  = difference / 1000;
    usec = (int)(difference * 1000) % 1000000;
  } else {
    sec  = 0;
    usec = 1;
  }

  /* GRONK */
  sec = 0;
  usec = 10;

  value.it_interval.tv_sec  = 0;
  value.it_interval.tv_usec = 0;
  value.it_value.tv_sec     = sec;
  value.it_value.tv_usec    = usec;

  if (setitimer(ITIMER_REAL, &value, NULL) == -1) {
    log("setitimer: %s\n", strerror(errno));
  }
}


/* cancel the pending callback set for t */
void
deschedule_callback (double time,
		     void *context)
{
  struct itimerval value;
  disksim_callback_fn = NULL;

  if (debug) log("descheduling callback %f\n", time);

  value.it_interval.tv_sec  = 0;
  value.it_interval.tv_usec = 0;
  value.it_value.tv_sec     = 0;
  value.it_value.tv_usec    = 0;

  if (setitimer(ITIMER_REAL, &value, NULL) == -1) {
    log("setitimer: %s\n", strerror(errno));
  }
}



static void
start_simulation (void)
{
  time_stamp(&ts_simulation_start);

  disksim_interface =
    disksim_interface_initialize("parfile",
				 "outfile",
				 handle_request_complete,
				 schedule_callback,
				 deschedule_callback,
				 0);
}


static void
stop_simulation (void)
{
  disksim_interface_shutdown(disksim_interface, get_simtime_msec());
}


/************************************************************************
 * Event processing
 ************************************************************************/

static void
handle_eof (int sock)
{
  const int max_events=1;
  struct kevent events[max_events];
  int n=0;

  assert(sock==memulator_sock);

  EV_SET(&events[n++], sock, EVFILT_READ, EV_DELETE, 0,0,0);
  assert(n<=max_events);
  if (kevent(kq_fd, events, n, NULL, 0, NULL) == -1) {
    log("kevent: %s\n", strerror(errno));
    assert(0);
  }

  if (sock == memulator_sock) {
    if (close(memulator_sock) == -1) {
      log("close: %s\n", strerror(errno));
    }
    memulator_sock = -1;
    received_termination = 1;
  }
  else {
    log("unknown socket %d\n", sock);
  }
}


static void
handle_ipc (int sock)
{
  ipc_message_t msg_in;

  assert(sock==memulator_sock);

  memset(&msg_in, 0, sizeof(ipc_message_t));
  recv_ipc(sock, &msg_in, sizeof(ipc_message_t));

  if (msg_in.magic != IPC_MAGIC) {
    log("bad magic %#x\n", msg_in.magic);
    return;
  }

  if (msg_in.source != IPC_SOURCE_MEMULATOR) {
    log("bad source %#x\n", msg_in.source);
    return;
  }

  switch (msg_in.type) {

  case IPC_TYPE_READ:
  case IPC_TYPE_WRITE:
    handle_request_arrive(&msg_in);
    break;

  case IPC_TYPE_DISCONNECT:
    log("%s\n", "Received disconnection directive.");
    handle_eof(sock);
    break;

  default:
    log("unknown type %d\n", msg_in.type);
  }
}


/************************************************************************
 * Main loop
 ************************************************************************/

static int
main_loop (void)
{
  const int max_events=16;
  struct kevent events[max_events];
  int n=0;

  /* Register kqueue for event notification */
  if ((kq_fd = kqueue()) == -1) {
    log("kqueue: %s\n", strerror(errno));
    assert(0);
  }

  signal(SIGHUP, SIG_IGN);
  signal(SIGINT, SIG_IGN);
  signal(SIGTERM, SIG_IGN);
  signal(SIGPROF, SIG_IGN);
  signal(SIGALRM, SIG_IGN);
  signal(SIGSTOP, SIG_IGN);
  signal(SIGTSTP, SIG_IGN);

  /* Set up some default events */
  //EV_SET(&events[n++], SIGHUP,  EVFILT_SIGNAL, EV_ADD, 0,0,0);
  //EV_SET(&events[n++], SIGINT,  EVFILT_SIGNAL, EV_ADD, 0,0,0);
  //EV_SET(&events[n++], SIGTERM, EVFILT_SIGNAL, EV_ADD, 0,0,0);
  EV_SET(&events[n++], SIGALRM, EVFILT_SIGNAL, EV_ADD, 0,0,0);
  if (memulator_sock != -1) {
    EV_SET(&events[n++], memulator_sock, EVFILT_READ, EV_ADD, 0,0,0);
  }

  assert(n<=max_events);
  if (kevent(kq_fd, events, n, NULL, 0, NULL) == -1) {
    log("kevent 1: %s\n", strerror(errno));
    assert(0);
  }

 while (!received_termination) {
    int num_events, i;

    /* Check for the next signal or read ready */
    /* GRONK: example code uses clever value for timeout */

    num_events = kevent(kq_fd, NULL, 0, events, max_events, NULL);
    if (num_events == -1) {
      if (errno == EINTR) continue;
      log("kevent 2: %s\n", strerror(errno));
      assert(0);
    }
    if (num_events > max_events) {
      log("kevent 3: %s\n", strerror(errno));
      assert(0);
    }

    /* Process all received events */

    for (i=0; i<num_events; i++) {
      if ((events[i].flags & EV_ERROR) != 0) {
	log("kevent 4: %s\n", strerror(errno));
	assert(0);
      }

      switch (events[i].filter) {

      case EVFILT_READ:
	if (events[i].ident == memulator_sock) {
	  if (!(events[i].flags & EV_EOF)) {
	    handle_ipc(events[i].ident);
	  }
	  else {  /* Other side disconnected */
	    handle_eof(events[i].ident);
	  }
	}
	else {
	  log("EVFILT_READ: unknown ident %#x\n", events[i].ident);
	}
	break;

      case EVFILT_SIGNAL:
	if (debug) log("signal received (%#x)\n", events[i].ident);
	if (events[i].ident == SIGALRM) {
	  if (disksim_callback_fn != 0) {
	    disksim_interface_callback_t callback = disksim_callback_fn;
	    disksim_callback_fn = NULL;
	    callback(disksim_interface, get_simtime_msec(), 0);
	  }
	  else {
	    log("%s\n", "Alarm but no callback registered!");
	  }
	}
	else {
	  log("Unknown signal received: %#x\n", events[i].ident);
	}
	break;

      default:
	log("unknown event %#x\n", events[i].filter);
	break;
      }
    }
  }

  return (0);  /* GRONK */
}


/************************************************************************
 * Main routines
 ************************************************************************/

static void
process_args (int argc, char **argv)
{
  char *program = argv[0];
  int ch;

  while ((ch = getopt(argc, argv, "df:")) != -1) {
    switch (ch) {
    case 'd':
      debug=1;
      break;
    case 'f':
      calibration_prefix = optarg;
      break;
    default:
      goto usage;
    }
  }
  argc -= optind;
  argv += optind;

  if (argc != 0) {
  usage:
    printf("\tUsage: %s [-d] [-f calibration_prefix]\n"
	   "\t\t-d: debug\n"
	   , program);
    assert(0);
  }
}


int
main (int argc, char **argv)
{
  int ret=0;

  process_args(argc, argv);

  open_calibration_times_file();

  start_simulation();

  memulator_attach();

  ret = main_loop();

  stop_simulation();

  close_calibration_times_file();

  return (ret);
}
