/*
 * The "Memulator" timing-accurate storage emulation environment
 * Authors: John Linwood Griffin, John Bucy, Greg Ganger
 *
 * Copyright (c) of Carnegie Mellon University, 2001-2004.
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <aio.h>
#include <errno.h>
#include <fcntl.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/event.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/shm.h>
#include <sys/mman.h>

#include "memulator.h"


int debug=0;
char *volume_filename=NULL;  /* Location of backing store file */
int volume_size_blocks=0;  /* Size of backing store in blocks */
int volume_fd=-1;  /* Descriptor for backing store */
int kq_fd=-1;  /* Descriptor for kernel event queue */
int memulator_sock=-1;  /* Socket for IPC with memulator */
int received_termination=0;  /* When memulator tells us to shut down */
char *sharedmem_file;
char *memory_preload_file = NULL;  /* File for initial memory load */
int memory_size_blocks=0;  /* Size of local memory buffer to create */
u_char *memory_buffer=MAP_FAILED;  /* Descriptor for local memory */
char *trace_dirname = NULL;  /* Directory for traced block requests */


/************************************************************************
 * Logging
 ************************************************************************/

#define log(a,b...) \
   do { fprintf(stdout, "%s: %s: " a, "store", __func__, b); } while (0)


void
print_sector(u_int8_t *p)
{
  int i;
  printf("store: ");
  for (i=0;i<512;i++) {
    printf("%02x ", p[i]);
    if (!((i+1)%16)) printf("\nstore: ");
  }
  printf("--------------------------------------------------------------\n");
}

/************************************************************************
 * Request tracking
 ************************************************************************/

typedef struct descriptor {
  TAILQ_ENTRY(descriptor) tqe;
  int64_t  offset_bytes;
  int      length_bytes;
  int      serial;
  int      type;
  u_int8_t *buffer;
  struct aiocb aiocb;
} descriptor_t;

TAILQ_HEAD(, descriptor) free_descrs = TAILQ_HEAD_INITIALIZER(free_descrs);

int num_free_descrs=0;
const int min_free_descrs=16;

descriptor_t *
get_free_descr (void)
{
  descriptor_t *descr;

  /* First check if we've reached the low-water mark for free descrs */

  if (num_free_descrs < min_free_descrs) {
    int i;
    for (i=0; i<min_free_descrs; i++) {
      descriptor_t *new_descr;

      while ((new_descr = malloc(sizeof(descriptor_t))) == NULL) {
	log("malloc: %s (retrying)\n", strerror(errno));
	sleep(1);
      }

      TAILQ_INSERT_TAIL(&free_descrs, new_descr, tqe);
      num_free_descrs++;
    }
  }

  assert(!TAILQ_EMPTY(&free_descrs));

  /* Grab the next available entry from the list of free descrs. */

  descr = TAILQ_FIRST(&free_descrs);
  TAILQ_REMOVE(&free_descrs, descr, tqe);
  num_free_descrs--;

  /* Zero-fill the new descriptor */
  memset(descr, 0, sizeof(descriptor_t));
  return (descr);
}


void
release_descr (descriptor_t *descr)
{
  TAILQ_INSERT_TAIL(&free_descrs, descr, tqe);
  num_free_descrs++;
}


/************************************************************************
 * Buffer management
 ************************************************************************/

u_int8_t *
get_buffer (int serial, int size)
{
  int shmid;
  u_int8_t *buffer;

  if ((shmid = shmget(serial, size, 0666)) == -1) {
    log("shmget: %s\n", strerror(errno));
    assert(0);
  }

  if ((buffer = shmat(shmid, NULL, 0)) == (void *)-1) {
    log("shmat: %s\n", strerror(errno));
    assert(0);
  }

  return (buffer);
}


void
release_buffer (u_int8_t *buffer)
{
  if (!buffer) return;

  if (shmdt(buffer) == -1) {
    log("shmdt: %s\n", strerror(errno));
  }
}


/************************************************************************
 * Interprocess communication
 ************************************************************************/

int
send_ipc (int sock, void *buf, size_t len)
{
  u_int8_t *msg;
  int bytes = 0;	/* Total bytes sent */
  int rc;

  if (debug) log("%s\n","send_ipc");

  msg = buf;

  while (bytes < len) {
    if ((rc = send(sock, msg + bytes, len - bytes, 0)) == -1) {
      log("send: %s", strerror(errno));
      return (-1);
    }
    bytes += rc;
  }

  if (bytes != len) {
    log("bad transfer (%d != %d)\n", bytes, len);
  }

  return (bytes);
}


int
recv_ipc (int sock, void *buf, size_t len)
{
  u_int8_t *msg;
  int bytes = 0;	/* Total bytes received */
  int rc;

  if (debug) log("%s\n","recv_ipc");

  msg = buf;

  while (bytes < len) {
    if ((rc = recv(sock, buf + bytes, len - bytes, 0)) == -1) {
      log("recv: %s", strerror(errno));
      return (-1);
    }
    bytes += rc;
  }

  if (bytes != len) {
    log("bad transfer (%d != %d)\n", bytes, len);
  }

  return (bytes);
}


void
memulator_attach (void)
{
  struct sockaddr_un sun;
  size_t size;
  ipc_message_t msg1_out, msg2_out;

  assert(memulator_sock==-1);

  /* socket */

  if ((memulator_sock = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) {
    log("socket: %s", strerror(errno));
    assert(0);
  }

  /* connect */

  memset(&sun, 0, sizeof(struct sockaddr_un));
  sun.sun_family = AF_UNIX;
  strncpy(sun.sun_path, MEMULATOR_SOCKET_FILENAME, 104);
  size = SUN_LEN(&sun);
  if (connect(memulator_sock, (struct sockaddr *)&sun, size) == -1) {
    log("connect: %s\n", strerror(errno));
    assert(0);
  }

  /* Send connect message */

  memset(&msg1_out, 0, sizeof(ipc_message_t));
  msg1_out.magic  = IPC_MAGIC;
  msg1_out.source = IPC_SOURCE_STORE;
  msg1_out.type   = IPC_TYPE_CONNECT;
  send_ipc(memulator_sock, &msg1_out, sizeof(ipc_message_t));

  /* Send volume size message */

  memset(&msg2_out, 0, sizeof(ipc_message_t));
  msg2_out.magic  = IPC_MAGIC;
  msg2_out.source = IPC_SOURCE_STORE;
  msg2_out.type   = IPC_TYPE_PARAMETERS;
  msg2_out.request_length_blocks = volume_size_blocks;
  send_ipc(memulator_sock, &msg2_out, sizeof(ipc_message_t));
}


/************************************************************************
 * Volume management
 ************************************************************************/

void
open_volume (void)
{
  struct stat stat;

  if (!volume_filename) return;

  assert(volume_fd == -1);

  if ((volume_fd = open(volume_filename, O_RDWR)) == -1) {
    log("open: %s\n", strerror(errno));
    assert(0);
  }

  memset(&stat, 0, sizeof(struct stat));
  if (fstat(volume_fd, &stat) == -1) {
    log("stat %s failed: %s\n", volume_filename, strerror(errno));
  }

  if (stat.st_size != 0) {
    volume_size_blocks = stat.st_size / 512;
  } else {
    if (!strcmp(volume_filename,"/dev/da1")) {
      volume_size_blocks = 35566478;
    }
  }
}


void
close_volume (void)
{
  if (volume_fd == -1) return;

  if (close(volume_fd) == -1) {
    log("close: %s\n", strerror(errno));
  }
  volume_fd=-1;
}


void
map_memory (void)
{
  if (!memory_size_blocks) return;

  if (volume_size_blocks) {
    log("%s\n", "Warning: volume_size_blocks already has a value.\n");
  }
  volume_size_blocks = memory_size_blocks;

  if (memory_preload_file) {
    int memory_preload_fd = -1;  /* File descriptor for initial memory load */
    int i;
    char deadbuf[512];

    memory_preload_fd = open(memory_preload_file, O_RDONLY);
    if (memory_preload_fd == -1) {
      log("Cannot open file %s: %s\n", memory_preload_file, strerror(errno));
      exit(1);
    }
    memory_buffer = mmap(NULL,
			 memory_size_blocks * 512,
			 PROT_READ | PROT_WRITE,
			 MAP_PRIVATE,
			 memory_preload_fd,
			 0);
    close(memory_preload_fd);

    if (memory_buffer == MAP_FAILED) {
      log("%s\n", strerror(errno));
      exit(1);
    }

    /* Make sure the memory is loaded */
    for (i = 0; i < memory_size_blocks; i++) {
      memcpy(deadbuf, memory_buffer + (512 * i), 512);
    }

  } else {
    memory_buffer = mmap(NULL,
			 memory_size_blocks * 512,
			 PROT_READ | PROT_WRITE,
			 MAP_ANON | MAP_PRIVATE,
			 -1,
			 0);

    if (memory_buffer == MAP_FAILED) {
      log("%s\n", strerror(errno));
      exit(1);
    }

    memset(memory_buffer, 0, memory_size_blocks * 512);
  }
}


void
unmap_memory (void)
{
  int rc;

  if (memory_buffer == MAP_FAILED) return;

  rc = munmap(memory_buffer, memory_size_blocks * 512);
  if (!rc) {
    log("%s\n", strerror(errno));
  }
}


/************************************************************************
 * Event processing
 ************************************************************************/

void
finish_request (struct descriptor *descr)
{
  ipc_message_t msg_out;

  /* If we're tracing all the block requests (including the data) for
   * a run, do so here. */
  if (trace_dirname) {
    int rc;
    int fd;
    char type='u';
    int filename_len = 512;
    char filename[filename_len];

    if (descr->type == IPC_TYPE_READ) type = 'r';
    if (descr->type == IPC_TYPE_WRITE) type = 'w';

    rc = snprintf(filename, filename_len,
		  "%s/%06d-%c-%lld-%d",
		  trace_dirname,
		  descr->serial,
		  type,
		  descr->offset_bytes / 512,
		  descr->length_bytes / 512);
    assert(rc >= 0);

    fd = open(filename, O_WRONLY|O_EXCL|O_FSYNC|O_DIRECT|O_CREAT, 0777);
    if (fd < 0) {
      log("cannot open %s: %s\n", filename, strerror(errno));
      trace_dirname = NULL;
    } else {
      rc = write(fd, descr->buffer, descr->length_bytes);
      if (rc != descr->length_bytes) {
	log("cannot write %s (%d != %d): %s\n", filename, rc, descr->length_bytes, strerror(errno));
	trace_dirname = NULL;
      }
      close(fd);
    }
  }

  release_buffer(descr->buffer);

  memset(&msg_out, 0, sizeof(ipc_message_t));
  msg_out.magic  = IPC_MAGIC;
  msg_out.source = IPC_SOURCE_STORE;
  msg_out.type   = IPC_TYPE_REQUEST_OK;
  msg_out.serial = descr->serial;
  send_ipc(memulator_sock, &msg_out, sizeof(ipc_message_t));

  release_descr(descr);
}



void
handle_eof (int sock)
{
  const int max_events=1;
  struct kevent events[max_events];
  int n=0;

  assert(sock==memulator_sock);

  EV_SET(&events[n++], sock, EVFILT_READ, EV_DELETE, 0,0,0);
  assert(n<=max_events);
  if (kevent(kq_fd, events, n, NULL, 0, NULL) == -1) {
    log("kevent: %s\n", strerror(errno));
    assert(0);
  }

  if (sock == memulator_sock) {
    if (close(memulator_sock) == -1) {
      log("close: %s\n", strerror(errno));
    }
    memulator_sock = -1;
    received_termination = 1;
  }
  else {
    log("unknown socket %d\n", sock);
  }
}


void
handle_ipc (int sock)
{
  ipc_message_t msg_in;

  assert(sock==memulator_sock);

  memset(&msg_in, 0, sizeof(ipc_message_t));
  recv_ipc(sock, &msg_in, sizeof(ipc_message_t));

  if (msg_in.magic != IPC_MAGIC) {
    log("bad magic %#x\n", msg_in.magic);
    return;
  }

  if (msg_in.source != IPC_SOURCE_MEMULATOR) {
    log("bad source %#x\n", msg_in.source);
    return;
  }

  switch (msg_in.type) {

  case IPC_TYPE_READ:
    {
      descriptor_t *descr;
      int rc;

      if (debug) log("request %d for block %lld (%d)\n", msg_in.serial, msg_in.request_offset_blocks, msg_in.request_length_blocks);

      descr = get_free_descr();
      descr->offset_bytes = msg_in.request_offset_blocks * 512;
      descr->length_bytes = msg_in.request_length_blocks * 512;
      descr->serial = msg_in.serial;
      descr->buffer = get_buffer(descr->serial, descr->length_bytes);
      descr->type = msg_in.type;

      assert((descr->offset_bytes % 512) == 0);

      /* Short-circuit if just using memory */
      if (memory_buffer != MAP_FAILED) {
	if ((descr->offset_bytes + descr->length_bytes)/512 > memory_size_blocks) {
	  log("%s\n", "You must increase memory_size_blocks.\n");
	  exit(1);
	}
	memcpy(descr->buffer, 
	       memory_buffer + descr->offset_bytes, 
	       descr->length_bytes);
	finish_request(descr);  descr = NULL;
	return;
      }

      descr->aiocb.aio_fildes = volume_fd;
      descr->aiocb.aio_offset = descr->offset_bytes;
      descr->aiocb.aio_buf    = descr->buffer;
      descr->aiocb.aio_nbytes = descr->length_bytes;
      descr->aiocb.aio_sigevent.sigev_notify        = SIGEV_KEVENT;
      descr->aiocb.aio_sigevent.sigev_notify_kqueue = kq_fd;
      descr->aiocb.aio_sigevent.sigev_value.sigval_ptr = descr;

      if (descr->type == IPC_TYPE_READ) {
	if ((rc = aio_read(&descr->aiocb)) == 0) {
	  /* Asynchronous I/O successfully started; nothing more to do */
	  return;
	}
	/* The asynchronous I/O call failed */
	log("aio_write: %s\n", strerror(errno));
      } 
      else if (descr->type == IPC_TYPE_WRITE) {
	// print_sector(descr->buffer);
	if ((rc = aio_write(&descr->aiocb)) == 0) {
	  /* Asynchronous I/O successfully started; nothing more to do */
	  return;
	}
	/* The asynchronous I/O call failed */
	log("aio_write: %s\n", strerror(errno));
      }
      else {
	assert(0);
      }
    }
    log("%s\n","GRONK: I shouldn't be here.");
    break;

  case IPC_TYPE_WRITE:
    /* Because the FreeBSD Asynchronous I/O (AIO) facility does not
     * correctly handle the case of read-immediately-following-write,
     * we must make all writes synchronous, for now.  This might change
     * once we put some caching code in. */
    {
      descriptor_t *descr;
      int rc;
      ipc_message_t msg_out;

      if (debug) log("(sync) request %d for block %lld (%d)\n", msg_in.serial, msg_in.request_offset_blocks, msg_in.request_length_blocks);

      descr = get_free_descr();
      descr->offset_bytes = msg_in.request_offset_blocks * 512;
      descr->length_bytes = msg_in.request_length_blocks * 512;
      descr->serial = msg_in.serial;
      descr->buffer = get_buffer(descr->serial, descr->length_bytes);
      descr->type = msg_in.type;

      assert((descr->offset_bytes % 512) == 0);

      /* Short-circuit if just using memory */
      if (memory_buffer != MAP_FAILED) {
	if ((descr->offset_bytes + descr->length_bytes)/512 > memory_size_blocks) {
	  log("%s\n", "You must increase memory_size_blocks.\n");
	  exit(1);
	}
	memcpy(memory_buffer + descr->offset_bytes, 
	       descr->buffer, 
	       descr->length_bytes);
	finish_request(descr);  descr = NULL;
	return;
      }

      rc = pwrite(volume_fd, descr->buffer,
		  descr->length_bytes, descr->offset_bytes);
      if (rc != descr->length_bytes) {
	log("pwrite: %s\n", strerror(errno));
      }

      release_buffer(descr->buffer); descr->buffer = NULL;

      memset(&msg_out, 0, sizeof(ipc_message_t));
      msg_out.magic  = IPC_MAGIC;
      msg_out.source = IPC_SOURCE_STORE;
      msg_out.type   = IPC_TYPE_REQUEST_OK;
      msg_out.serial = descr->serial;
      send_ipc(memulator_sock, &msg_out, sizeof(ipc_message_t));

      release_descr(descr); descr = NULL;
      return;
    }
    log("%s\n","GRONK: I shouldn't be here.");
    break;

  case IPC_TYPE_DISCONNECT:
    log("%s\n", "Received disconnection directive.");
    handle_eof(sock);
    break;

  default:
    log("unknown type %d\n", msg_in.type);
  }
}


void
handle_aio (descriptor_t *descr)
{
  int rc;

  if (debug) log("aio done for %d\n", descr->serial);

  if ((rc = aio_return(&descr->aiocb)) != descr->length_bytes) {
    log("aio_return returned %d != %d\n", rc, descr->length_bytes);
  }

  finish_request(descr);  descr = NULL;
}


/************************************************************************
 * Main loop
 ************************************************************************/

int
main_loop (void)
{
  const int max_events=16;
  struct kevent events[max_events];
  int n=0;

  /* Register kqueue for event notification */
  if ((kq_fd = kqueue()) == -1) {
    log("kqueue: %s\n", strerror(errno));
    assert(0);
  }

  signal(SIGHUP, SIG_IGN);
  signal(SIGINT, SIG_IGN);
  signal(SIGTERM, SIG_IGN);
  signal(SIGPROF, SIG_IGN);
  signal(SIGALRM, SIG_IGN);
  signal(SIGSTOP, SIG_IGN);
  signal(SIGTSTP, SIG_IGN);

  /* Set up some default events */
  //EV_SET(&events[n++], SIGHUP,  EVFILT_SIGNAL, EV_ADD, 0,0,0);
  //EV_SET(&events[n++], SIGINT,  EVFILT_SIGNAL, EV_ADD, 0,0,0);
  //EV_SET(&events[n++], SIGTERM, EVFILT_SIGNAL, EV_ADD, 0,0,0);
  if (memulator_sock != -1) {
    EV_SET(&events[n++], memulator_sock, EVFILT_READ, EV_ADD, 0,0,0);
  }

  assert(n<=max_events);
  if (kevent(kq_fd, events, n, NULL, 0, NULL) == -1) {
    log("kevent 1: %s\n", strerror(errno));
    assert(0);
  }

  while (!received_termination) {
    int num_events, i;

    /* Check for the next signal or read ready */
    /* GRONK: example code uses clever value for timeout */

    num_events = kevent(kq_fd, NULL, 0, events, max_events, NULL);
    if (num_events == -1) {
      if (errno == EINTR) continue;
      log("kevent 2: %s\n", strerror(errno));
      assert(0);
    }
    if (num_events > max_events) {
      log("kevent 3: %s\n", strerror(errno));
      assert(0);
    }

    /* Process all received events */

    for (i=0; i<num_events; i++) {
      if ((events[i].flags & EV_ERROR) != 0) {
	log("kevent 4: %s\n", strerror(errno));
	assert(0);
      }

      switch (events[i].filter) {

      case EVFILT_READ:
	if (events[i].ident == memulator_sock) {
	  if (!(events[i].flags & EV_EOF)) {
	    handle_ipc(events[i].ident);
	  }
	  else {  /* Other side disconnected */
	    handle_eof(events[i].ident);
	  }
	}
	else {
	  log("EVFILT_READ: unknown ident %#x\n", events[i].ident);
	}
	break;

      case EVFILT_AIO:
	handle_aio((descriptor_t *)events[i].udata);
	break;

      default:
	log("unknown event %#x\n", events[i].filter);
	break;
      }
    }
  }

  return (0);  /* GRONK */
}


/************************************************************************
 * supplementary stuff
 ************************************************************************/

void 
make_trace_dir (void)
{
  int rc;

  if (!trace_dirname) return;

  rc = mkdir(trace_dirname, 0777);
  if (rc < 0) {
    log("cannot mkdir %s: %s\n", trace_dirname, strerror(errno));
    trace_dirname = NULL;
  }
}


/************************************************************************
 * Main routines
 ************************************************************************/

void
process_args (int argc, char **argv)
{
  char *program = argv[0];
  int ch;

  while ((ch = getopt(argc, argv, "b:dr:x:y:")) != -1) {
    switch (ch) {
    case 'b':
      volume_filename = optarg;
      break;
    case 'd':
      debug=1;
      break;
    case 'r':
      trace_dirname = optarg;
      break;
    case 'x':
      memory_preload_file = optarg;
      break;
    case 'y':
      memory_size_blocks = atoi(optarg);
      break;
    default:
      goto usage;
    }
  }
  argc -= optind;
  argv += optind;

  if (argc != 0
      || (!volume_filename && !memory_size_blocks)
      || (volume_filename && memory_size_blocks)) {
  usage:
    printf("Usage: %s [-b filename] [-d] [-y size]\n"
	   "  -b file: what to use for the backing store (e.g., /dev/da2)\n"
	   "  -d: debug\n"
	   "  -x file: the file to preload the memory with\n"
	   "  -y size: amount of memory to allocate for local store\n"
	   "  ... the -b and -y options are mutually exclusive\n"
	   , program);
    assert(0);
    exit(1);
  }
}


int
main (int argc, char **argv)
{
  int ret=0;

  process_args(argc, argv);

  log("%s\n", "store starting");

  make_trace_dir();

  open_volume();
  map_memory();

  memulator_attach();

  log("%s\n", "store ready");

  ret = main_loop();

  unmap_memory();
  close_volume();

  return (ret);
}
