/*
 * The "Memulator" timing-accurate storage emulation environment
 * Authors: John Linwood Griffin, John Bucy, Greg Ganger
 *
 * Copyright (c) of Carnegie Mellon University, 2001-2004.
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 *
 */

#define MEMULATOR_EVENT_HOST_KERNEL_ARRIVE   (0x21)
#define MEMULATOR_EVENT_HOST_KERNEL_COMPLETE (0x22)

#include <linux/kernel.h>
#include <linux/module.h>

#include <linux/sched.h>
#include <asm/uaccess.h>  /* get_fs, set_fs */
#include <sys/syscall.h>
#include <asm/fcntl.h>

/* GRONK: fix the include path to include these directories? */
#include <linux/blkdev.h>
#include "/usr/src/linux/drivers/scsi/scsi.h"
#include "/usr/src/linux/drivers/scsi/hosts.h"

const char *NAME = "Hosttimes module";

static char *filename = NULL;    /* Output filename for timestamps */
static int host_no = -1;  /* Host adapter connecting to watched device */
static int channel = -1;  /* Channel number of watched device */
static int target  = -1;  /* Target id of watched device */
static int lun     = -1;  /* LUN of watched device */
MODULE_PARM(filename, "s");
MODULE_PARM(host_no, "i");
MODULE_PARM(channel, "i");
MODULE_PARM(target, "i");
MODULE_PARM(lun, "i");

static char output_ready = 0;  /* Set when the daemon can accept output. */
static char module_done  = 0;  /* Set when the module is ready to unload */
static char kthread_done = 0;  /* Set when the kernel thread completes. */

static struct semaphore buffer_sema;
/* buffer_sema: Global lock for coordinating updates to buffers. */
DECLARE_WAIT_QUEUE_HEAD(kthread_waitq);
/* kthread_waitq: used to wake up the kernel thread that writes the
 * buffers out to disk. */

extern spinlock_t io_request_lock;  /* defined in linux/blk.h */

#define BUFFER_DATA_SIZE (262144)  /* it's okay to increase this */
static struct new_buffer {
  int bytes_available;
  int bytes_allocated;
  u_char data[BUFFER_DATA_SIZE];
} new_buf1, new_buf2, *active_buf = &new_buf1, *inactive_buf = &new_buf2;

static int num_buf_switches = 0;
static int max_bytes_allocated = -1;
static int min_bytes_available = -1;


/* This is used to keep the original function pointers for after we've
 * interposed our own function pointers for the "queuecommand" and
 * "done" functions. */
int (*original_queuecommand)(Scsi_Cmnd *, void (*done)(Scsi_Cmnd *)) = NULL;
void (*original_done)(Scsi_Cmnd *) = NULL;


static inline void
time_stamp (u_int64_t *d)
{
  asm volatile ("rdtsc" : "=&A" (*d));
}


static int
hosttimesd (void *arg)
{
  /* Function pointers to the system calls necessary for file output */
  extern void *sys_call_table[];
  asmlinkage long (*syscall_chmod)
    (const char *, mode_t) = sys_call_table[SYS_chmod];
  asmlinkage long (*syscall_chown)
    (const char *, uid_t, gid_t) = sys_call_table[SYS_chown];
  asmlinkage long (*syscall_close)
    (unsigned int) = sys_call_table[SYS_close];
  asmlinkage long (*syscall_open)
    (const char *, int, mode_t) = sys_call_table[SYS_open];
  asmlinkage ssize_t (*syscall_write)
    (unsigned int, const char *, size_t) = sys_call_table[SYS_write];

  /* Local variables */
  mm_segment_t fs;  /* Need to change the segment pointers for file output */
  int output_fd;

  /* Kernel thread initialization */
  daemonize();
  sigfillset(&current->blocked);
  sprintf(current->comm, __func__);

  /* Adjust the memory area.  This is necessary in order to successfully
   * invoke the syscalls. */
  fs = get_fs();
  set_fs(get_ds());

  /* Open the output file, or die trying. */
  output_fd = syscall_open(filename, O_CREAT|O_APPEND|O_WRONLY, 0666);
  if (output_fd < 0) {
    output_ready = 0;
    printk("%s: WARNING: cannot open %s: %d.\n", NAME, filename, output_fd);
    goto end;
  }
  syscall_chown(filename, 377, 172);  /* user=griffin2, gid=lcs */
  syscall_chmod(filename, 0666);      /* mode=-rw-rw-rw- */

  /* Allocate the buffers */
  active_buf->bytes_available = BUFFER_DATA_SIZE - 1;
  active_buf->bytes_allocated = 0;

  inactive_buf->bytes_available = BUFFER_DATA_SIZE - 1;
  inactive_buf->bytes_allocated = 0;

  /* At initialization time we hold the buffer lock by fiat.  Release it
   * here to allow the writers to start. */
  output_ready = 1;
  up(&buffer_sema);

  for (;;) {

    /* What this function does:
     *
     * We are woken up by the clock, or a high-water alarm.
     * If the buffer is empty, we fall back asleep (end).
     * We take the lock.
     * We switch the main buffer pointer to buffer B.
     * We release the lock.
     * If there's anything in buffer A, we flush buffer A.
     *
     * What the writers do:
     *
     * They take the lock.
     * They check whether or not they're allowed to write.
     * They do some bounds-checking.
     * They write into the buffer.
     * They update the buffer pointers.
     * They release the lock.
     */

    /* Set an alarm to wake us up once per hundredth second (this may happen more
     * quickly if someone wants to notify us we're running out of buffer
     * space). */
    if (!module_done) interruptible_sleep_on_timeout(&kthread_waitq, HZ);

    /* If we're awake but there's nothing to do, go back to sleep. */
    if (!module_done && (active_buf->bytes_allocated == 0)) continue;

    /* Lock */
    down(&buffer_sema);

    if (module_done) output_ready = 0;

    /* Swap the active and inactive buffers, so we can immediately
     * release the semaphore. */
    if (active_buf->bytes_allocated != 0) {
      struct new_buffer *temp_buf = active_buf;
      active_buf = inactive_buf;
      inactive_buf = temp_buf;
    }

    /* Unlock */
    up(&buffer_sema);

    /* Now, if there is any data in the inactive buffer (what used to be
     * the active buffer), flush it to disk and reset the buffer. */
    if (inactive_buf->bytes_allocated != 0) {
      num_buf_switches++;

      if (min_bytes_available == -1
	  || inactive_buf->bytes_available < min_bytes_available) {
	min_bytes_available = inactive_buf->bytes_available;
      }

      if (max_bytes_allocated == -1
	  || inactive_buf->bytes_allocated > max_bytes_allocated) {
	max_bytes_allocated = inactive_buf->bytes_allocated;
      }

      syscall_write(output_fd,
		    inactive_buf->data, 
		    inactive_buf->bytes_allocated);
    }

    inactive_buf->bytes_available += inactive_buf->bytes_allocated;
    inactive_buf->bytes_allocated = 0;

    if (!output_ready) break;  /* We've received the module_done signal */
  }

  /* GRONK: print shutdown time and rdtsc count */

  syscall_close(output_fd);

 end:
  /* Restore the memory area to ensure the system stays in a consistent
   * state. */
  set_fs(fs);

  kthread_done = 1;
  return 0;
}


static void
new_done (Scsi_Cmnd *scp)
{
  unsigned int offset;
  unsigned int length;
  unsigned int opcode;
  unsigned int code;
  u_int64_t timestamp;
  int bytes;

  time_stamp(&timestamp);

  opcode = scp->cmnd[0];
  code = MEMULATOR_EVENT_HOST_KERNEL_COMPLETE;

  switch (scp->cmnd[0]) {
  case 0x28:	/* READ 10  */
  case 0x2a:	/* WRITE 10 */
    offset = scp->cmnd[2] << 24 
      | scp->cmnd[3] << 16 
      | scp->cmnd[4] << 8 
      | scp->cmnd[5];
    length = scp->cmnd[7] << 8 
      | scp->cmnd[8];
    break;

  case 0x08:	/* READ 6 */
  case 0x0a:	/* WRITE 6 */
    offset = (scp->cmnd[1] & 0x1f) << 16 
      | scp->cmnd[2] << 8 
      | scp->cmnd[3];
    length = scp->cmnd[4] ? scp->cmnd[4] : 256;
    break;

  default:
    offset = length = 0;
  }

  /* What this function does:
   *
   * We take the lock.
   * We check whether or not we're allowed to write.
   * We do some bounds-checking.
   * We write into the buffer and update the buffer pointers.
   * We release the lock.
   */

  /* Lock */
  down(&buffer_sema);

  if (!output_ready) {
    /* Unlock and continue execution as though we didn't exist. */
    goto end;
  }

  bytes = snprintf(active_buf->data + active_buf->bytes_allocated,
		   active_buf->bytes_available,
		   "%d\t%d\t%x\t%x\t%lld\n",
		   offset, length, opcode, code, timestamp);

  active_buf->bytes_allocated += bytes;
  active_buf->bytes_available -= bytes;

  if (active_buf->bytes_available <= 0) {
    /* Unlock and continue execution as though we didn't exist. */
    output_ready = 0;
    printk("%s: WARNING: you must increase buffer_data_size.\n", NAME);
    goto end;
  }

  /* Unlock and continue execution with the original done function. */
 end:
  up(&buffer_sema);
  MOD_DEC_USE_COUNT;
  original_done(scp);
}


static int
new_queuecommand (Scsi_Cmnd *scp, void (*done)(Scsi_Cmnd *))
{
  unsigned int offset;
  unsigned int length;
  unsigned int opcode;
  unsigned int code;
  u_int64_t timestamp;
  int bytes;

  /* Short-cut the common case */
  if (scp->host->host_no != host_no
      || scp->channel != channel
      || scp->target != target
      || scp->lun != lun) {
    return original_queuecommand(scp, done);
  }

  time_stamp(&timestamp);

  opcode = scp->cmnd[0];
  code = MEMULATOR_EVENT_HOST_KERNEL_ARRIVE;

  switch (scp->cmnd[0]) {
  case 0x28:	/* READ 10  */
  case 0x2a:	/* WRITE 10 */
    offset = scp->cmnd[2] << 24 
      | scp->cmnd[3] << 16 
      | scp->cmnd[4] << 8 
      | scp->cmnd[5];
    length = scp->cmnd[7] << 8 
      | scp->cmnd[8];
    break;

  case 0x08:	/* READ 6 */
  case 0x0a:	/* WRITE 6 */
    offset = (scp->cmnd[1] & 0x1f) << 16 
      | scp->cmnd[2] << 8 
      | scp->cmnd[3];
    length = scp->cmnd[4] ? scp->cmnd[4] : 256;
    break;

  default:
    offset = length = 0;
  }

  /* What this function does:
   *
   * We take the lock.
   * We check whether or not we're allowed to write.
   * We do some bounds-checking.
   * We write into the buffer.
   * We update the buffer pointers.
   * We release the lock.
   */

  /* Lock */
  down(&buffer_sema);

  if (!output_ready) {
    /* Unlock and continue execution as though we didn't exist. */
    up(&buffer_sema);
    return original_queuecommand(scp, done);
  }

  /* If this is the first time we've seen a request, make note of the
   * "done" function.  If we see a different "done" function than we've
   * seen before, panic. */
  if (!original_done) {
    original_done = done;
  } else if (original_done != done) {
    output_ready = 0;
    printk("%s: WARNING: done function mismatch.\n", NAME);
    up(&buffer_sema);
    return original_queuecommand(scp, done);
  }

  bytes = snprintf(active_buf->data + active_buf->bytes_allocated,
		   active_buf->bytes_available,
		   "%d\t%d\t%x\t%x\t%lld\n",
		   offset, length, opcode, code, timestamp);

  active_buf->bytes_allocated += bytes;
  active_buf->bytes_available -= bytes;

  if (active_buf->bytes_available <= 0) {
    /* Unlock and continue execution as though we didn't exist. */
    output_ready = 0;
    printk("%s: WARNING: you must increase buffer_data_size.\n", NAME);
    up(&buffer_sema);
    return original_queuecommand(scp, done);
  }

  /* Unlock */
  up(&buffer_sema);

  /* Send the request on its merry way. */
  MOD_INC_USE_COUNT;
  return original_queuecommand(scp, new_done);
}


int
init_module (void)
{
  if (!filename
      || host_no == -1
      || channel == -1
      || target == -1
      || lun == -1) {
    printk("%s: WARNING: Invalid parameters (%s, scsi%d:%d:%d:%d)\n", 
	   NAME, filename, host_no, channel, target, lun);
    return -EINVAL;
  }

  printk("%s: filename %s\n", NAME, filename);

  sema_init(&buffer_sema, 0);  /* Starts locked */

  /* Lock the kernel, then find the structures related to the module
   * parameters (or die trying), then insert the interposition code in
   * the critical path of the requests, then unlock the kernel. */
  {
    unsigned long flags;
    struct Scsi_Host *host;

    spin_lock_irqsave(&io_request_lock, flags);

    for (host = scsi_hostlist; 
	 host && host->host_no != host_no; 
	 host = host->next) {
      // printk("Notyet...\n");
    };
    if (host == NULL) {
      printk("%s: WARNING: unable to locate host_no %d\n", NAME, host_no);
      spin_unlock_irqrestore(&io_request_lock, flags);
      return -EINVAL;
    }

    original_queuecommand = host->hostt->queuecommand;
    host->hostt->queuecommand = new_queuecommand;

    printk("%s: installed for %s\n", NAME, host->hostt->name);

    spin_unlock_irqrestore(&io_request_lock, flags);
  }

  /* Start the kernel writer thread */
  kernel_thread(hosttimesd, NULL, 0);

  // printk("%s: loaded.\n", NAME);
  return 0;
}


void
cleanup_module (void)
{
  int i;
  DECLARE_WAIT_QUEUE_HEAD(cleanup_waitq);

  /* Lock the kernel, then find the structures related to the module
   * parameters, then remove the interposition code from the critical
   * path of requests, then unlock the kernel. */
  {
    unsigned long flags;
    struct Scsi_Host *host;

    spin_lock_irqsave(&io_request_lock, flags);

    for (host = scsi_hostlist; 
	 host && host->host_no != host_no; 
	 host = host->next);
    if (host != NULL) {
      host->hostt->queuecommand = original_queuecommand;
      original_queuecommand = NULL;
    } else {
      printk("%s: WARNING: unable to locate host_no %d\n", NAME, host_no);
    }

    spin_unlock_irqrestore(&io_request_lock, flags);
  }

  module_done = 1;

  /* Wait for up to 30 seconds before removing the module.  I'm not
   * really sure I've taken care of the kthread processing correctly --
   * is there a possible race condition on cleaning up the kthread
   * state? -- but this should work okay for experimental code. */

  // printk("%s: waiting for kernel thread to complete.\n", NAME);

  for (i = 0; (i < 30) && !kthread_done; i++) {
    sleep_on_timeout(&cleanup_waitq, HZ);
  }
  if (i==30) {
    // printk("...gave up.\n");
  }

  printk("%s: stats: num %d, min available %d, max allocated %d\n", NAME, 
	 num_buf_switches,
	 min_bytes_available,
	 max_bytes_allocated);

  printk("%s: unloaded.\n", NAME);
}
