/*
 * The "Memulator" timing-accurate storage emulation environment
 * Authors: John Linwood Griffin, John Bucy, Greg Ganger
 *
 * Copyright (c) of Carnegie Mellon University, 2001-2004.
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <assert.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <scsi/sg.h>
#include <string.h>
#include <sys/poll.h>

static char *prefix = NULL;
static int sg_num = -1;
static int sg_fd;

struct event {
  u_int32_t offset;
  u_int16_t length;
  u_int8_t  opcode;
  u_int8_t  code;
  u_int64_t timestamp;
};  /* 128 bits */

#define log(a,b...) \
  do { fprintf(stdout, "%s: %s: " a, "microgenerate", __func__, b); } while (0)


static void
send_scsi_command (int    dxfer_direction,
		   u_char cmd_len,
		   u_int  dxfer_len,
		   void   *dxferp,
		   u_char *cmdp,
		   int    pack_id)
{
  sg_io_hdr_t io_hdr;
  int rc;

  memset(&io_hdr, 0, sizeof(sg_io_hdr_t));

  /* int interface_id;		[i] 'S' for SCSI generic (required) */
  /* int dxfer_direction;	[i] data transfer direction  */
  /* unsigned char cmd_len;	[i] SCSI command length ( <= 16 bytes) */
  /* unsigned char mx_sb_len;	[i] max length to write to sbp */
  /* unsigned short iovec_count;	[i] 0 implies no scatter gather */
  /* unsigned int dxfer_len;	[i] byte count of data transfer */
  /* void * dxferp;		[i], [*io] points to data transfer memory
					   or scatter gather list */
  /* unsigned char * cmdp;	[i], [*i] points to command to perform */
  /* unsigned char * sbp;	[i], [*o] points to sense_buffer memory */
  /* unsigned int timeout;	[i] MAX_UINT->no timeout (unit: millisec) */
  /* unsigned int flags;	[i] 0 -> default, see SG_FLAG... */
  /* int pack_id;		[i->o] unused internally (normally) */
  /* void * usr_ptr;		[i->o] unused internally */

  io_hdr.interface_id    = 'S';
  io_hdr.dxfer_direction = dxfer_direction;
  io_hdr.cmd_len         = cmd_len;
  io_hdr.mx_sb_len       = 0;  /* GROK: add sense buffer */
  io_hdr.iovec_count     = 0;
  io_hdr.dxfer_len       = dxfer_len;
  io_hdr.dxferp          = dxferp;
  io_hdr.cmdp            = cmdp;
  io_hdr.sbp             = NULL;  /* GROK: add sense_buffer */
  io_hdr.timeout         = 20000;
  io_hdr.flags           = SG_FLAG_DIRECT_IO;  /* GROK? */
  io_hdr.pack_id         = pack_id;
  io_hdr.usr_ptr         = NULL;

  rc = write(sg_fd, &io_hdr, sizeof(sg_io_hdr_t));
  if (rc != sizeof(sg_io_hdr_t)) {
    log("write failed (%d != %d): %s\n", rc, sizeof(sg_io_hdr_t), strerror(errno));
    exit(-1);
  }
}


int
recv_completion (int pack_id)
{
  sg_io_hdr_t io_hdr;
  struct pollfd pollfd;
  int rc;

  pollfd.fd = sg_fd;
  pollfd.events = POLLPRI | POLLIN;
 retry:
  rc = poll(&pollfd, 1, -1);
  if (rc < 1) {
    if (rc == -1 && errno == EINTR) goto retry;
    log("poll failed (%d): %s\n", rc, strerror(errno));
    assert(0);
  }

  memset(&io_hdr, 0, sizeof(sg_io_hdr_t));
  io_hdr.pack_id = pack_id;
  io_hdr.dxfer_direction = SG_DXFER_NONE;

  /* int pack_id;		[i->o] unused internally (normally) */
  /* void * usr_ptr;		[i->o] unused internally */
  /* unsigned char status;	[o] scsi status */
  /* unsigned char masked_status;	[o] shifted, masked scsi status */
  /* unsigned char msg_status;	[o] messaging level data (optional) */
  /* unsigned char sb_len_wr;	[o] byte count actually written to sbp */
  /* unsigned short host_status;	[o] errors from host adapter */
  /* unsigned short driver_status;	[o] errors from software driver */
  /* int resid;			[o] dxfer_len - actual_transferred */
  /* unsigned int duration;	[o] time taken by cmd (unit: millisec) */
  /* unsigned int info;		[o] auxiliary information */

  rc = read(sg_fd, &io_hdr, sizeof(sg_io_hdr_t));
  assert(rc == sizeof(sg_io_hdr_t));

  return (io_hdr.pack_id);
}


void
send_0x12_inquiry (unsigned char *buf,
		   unsigned int buflen,
		   int pack_id)
{
  u_int8_t scsi_command[12];

  memset(scsi_command, 0, 12);
  scsi_command[0] = 0x12;   /* opcode */
  scsi_command[4] = buflen; /* allocation length */

  send_scsi_command(SG_DXFER_FROM_DEV,	/* dxfer_direction */
		    6,			/* cmd_len */
		    buflen,		/* dxfer_len */
		    buf,			/* dxferp */
		    scsi_command,		/* cmdp */
		    pack_id);		/* pack_id */
}


void
send_0x28_read_10 (u_char *buf,  /* buffer for the read data */
		   u_int buflen, /* length in bytes of buf */
		   u_int block,  /* starting LBN for read */
		   u_int length, /* number of blocks to read */
		   int pack_id)
{
  u_int8_t scsi_command[12];

  assert(buflen >= (length * 512));

  memset(scsi_command, 0, 12);
  scsi_command[0] = 0x28;	/* opcode */
  scsi_command[2] = (block & 0xff000000) >> 24;	/* logical block address msb */
  scsi_command[3] = (block & 0x00ff0000) >> 16;	/* logical block address     */
  scsi_command[4] = (block & 0x0000ff00) >> 8;	/* logical block address     */
  scsi_command[5] = (block & 0x000000ff);	/* logical block address lsb */
  scsi_command[7] = (length & 0xff00) >> 8;	/* transfer length msb */
  scsi_command[8] = (length & 0x00ff);		/* transfer length lsb */

  send_scsi_command(SG_DXFER_FROM_DEV,	/* dxfer_direction */
		    10,			/* cmd_len */
		    buflen,		/* dxfer_len */
		    buf,			/* dxferp */
		    scsi_command,		/* cmdp */
		    pack_id);		/* pack_id */
}


void
send_0x2a_write_10 (u_char *buf,  /* buffer with the write data */
		    u_int buflen, /* length in bytes of buf */
		    u_int block,  /* starting LBN for write */
		    u_int length, /* number of blocks to write */
		    int pack_id)
{
  u_int8_t scsi_command[12];

  assert(buflen >= length * 512);

  memset(scsi_command, 0, 12);
  scsi_command[0] = 0x2a;	/* opcode */
  scsi_command[2] = (block & 0xff000000) >> 24;	/* logical block address msb */
  scsi_command[3] = (block & 0x00ff0000) >> 16;	/* logical block address     */
  scsi_command[4] = (block & 0x0000ff00) >> 8;	/* logical block address     */
  scsi_command[5] = (block & 0x000000ff);	/* logical block address lsb */
  scsi_command[7] = (length & 0xff00) >> 8;	/* transfer length msb */
  scsi_command[8] = (length & 0x00ff);		/* transfer length lsb */

  send_scsi_command(SG_DXFER_TO_DEV,	/* dxfer_direction */
		    10,			/* cmd_len */
		    buflen,		/* dxfer_len */
		    buf,			/* dxferp */
		    scsi_command,		/* cmdp */
		    pack_id);		/* pack_id */
}

/************************************************************************/

static int next_pack_id = 1;

void
check_emulator (void)
{
  int pack_id = next_pack_id++;
  u_int buflen = 256;
  u_char buf[buflen];

  send_0x12_inquiry(buf, buflen, pack_id);
  recv_completion(pack_id);

  if (strncmp(buf + 8, "Carnegie", 8)
      || strncmp(buf + 16, "Mellon Memulator", 16)
      || strncmp(buf + 32, "2004", 4)) {
    log("Invalid device: %.8s %.16s %.4s\n", buf + 8, buf + 16, buf + 32);
    exit(1);
  }

  log("Inquiry: %.8s %.16s %.4s\n", buf + 8, buf + 16, buf + 32);
}


void
do_read (u_int offset, u_int length)
{
  int pack_id = next_pack_id++;
  u_int buflen = length * 512;
  u_char buf[buflen];

  send_0x28_read_10(buf, buflen, offset, length, pack_id);
  recv_completion(pack_id);
}

/************************************************************************/

void
open_sg (void)
{
  const int filename_maxlen = 80;
  char filename[filename_maxlen];
  int k = 0;
  const int one=1;

  assert(sg_num != -1);

  snprintf(filename, filename_maxlen, "/dev/sg%d", sg_num);
  sg_fd = open(filename, O_RDWR);
  if (sg_fd == -1) {
    log("unable to open %s: %s\n", filename, strerror(errno));
    assert(0);
  }

  if ((ioctl(sg_fd, SG_GET_VERSION_NUM, &k) < 0) || (k < 30000)) {
    log("ioctl %s: %s\n", "SG_GET_VERSION_NUM", strerror(errno));
    assert(0);
  }

  if (ioctl(sg_fd, SG_SET_FORCE_PACK_ID, &one) == -1) {
    log("ioctl %s: %s\n", "SG_SET_FORCE_PACK_ID", strerror(errno));
    assert(0);
  }
}


void
close_sg (void)
{
  if (sg_fd == -1) return;

  close(sg_fd); sg_fd = -1;
}


void
process_args (int argc, char **argv)
{
  char *program = argv[0];
  int ch;

  while ((ch = getopt(argc, argv, "f:g:")) != -1) {
    switch (ch) {
    case 'f':  /* output file prefix */
      prefix = optarg;
      break;
    case 'g':  /* scsi generic device number */
      sg_num = atoi(optarg);
      break;
    default:
      goto usage;
    }
  }
  argc -= optind;
  argv += optind;

  if (argc != 0) {
    printf("argc==%d\n", argc);
  usage:
    printf("usage:\n"
	   "\t%s -f output_prefix -g scsi_generic_device_number\n"
	   , program);
    exit(1);
  }
}



int
main (int argc, char **argv)
{
  process_args(argc, argv);

  open_sg();

  check_emulator();

  {
    int i;
    for (i=0; i<1000; i++) {
      int rand_offset = 1+(int) (17916239.0*rand()/(RAND_MAX+1.0));
      int rand_length = 1+(int) (127.0*rand()/(RAND_MAX+1.0));

      if (1) {
	do_read(rand_offset, rand_length);
      } else {
	do_read(8, 16);
      }
    }
  }

  close_sg();

  printf("done\n");

  return 0;
}
