
#define MEMULATOR_EVENT_HOST_KERNEL_ARRIVE   (0x21)
#define MEMULATOR_EVENT_HOST_KERNEL_COMPLETE (0x22)
#define MEMULATOR_EVENT_TARGET_EMULATOR_ARRIVE	 (0x31)
#define MEMULATOR_EVENT_TARGET_EMULATOR_COMPLETE (0x32)
#define MEMULATOR_EVENT_TARGET_SIMULATOR_TIME    (0x40)

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <assert.h>
#include <errno.h>
#include <string.h>
#include <malloc.h>
#include <stdlib.h>

static char *prefix = NULL;
static char *output_string = NULL;

struct event {
  u_int32_t offset;
  u_int16_t length;
  u_int8_t  opcode;
  u_int8_t  code;
  union {
    u_int64_t tsc;  /* timestamp counter value from RDTSC */
    double    elapsed;  /* elapsed time for a request */
  } u;
};  /* 128 bits */


struct timeinfo {
  u_int32_t offset;
  u_int16_t length;
  u_int8_t  opcode;
  u_int8_t  flags;
  u_int64_t hos_arrive;
  u_int64_t hos_complete;
  u_int64_t mem_arrive;
  u_int64_t mem_complete;
  double    sim_time;
};

struct timeinfo *ti_hos = NULL;
int ti_hos_num_entries = 0;

struct timeinfo *ti_mem = NULL;
int ti_mem_num_entries = 0;

struct timeinfo *ti_sim = NULL;
int ti_sim_num_entries = 0;

double hos_mem_avg = 0.0;
double hos_mem_min = 0.0;
double hos_mem_max = 0.0;
double hos_mem_stddev = 0.0;
double hos_sim_avg = 0.0;
double hos_sim_min = 0.0;
double hos_sim_max = 0.0;
double hos_sim_stddev = 0.0;
double mem_sim_avg = 0.0;
double mem_sim_min = 0.0;
double mem_sim_max = 0.0;
double mem_sim_stddev = 0.0;

#define log(a,b...) \
   do { fprintf(stdout, "%s: %s: " a, "analyze", __func__, b); } while (0)


void
process_hos (void)
{
  int fd;
  int num_events;
  struct event *ev_arrive, *ev_complete;
  int ev_arrive_index = 0, ev_complete_index = 0;

  /* Open the file, use the stat information to determine the length
   * of the file, then use that to determine the length of the
   * ev_arrive and ev_complete arrays that we need to allocate. */
  {
    const int filename_maxlen = 80;
    char filename[filename_maxlen];
    struct stat stat_buf;

    assert((strlen(prefix) + 4) < filename_maxlen);
    snprintf(filename, filename_maxlen, "%s.hos", prefix);

    fd = open(filename, O_RDONLY);
    if (fd == -1) {
      log("unable to open %s: %s\n", filename, strerror(errno));
      assert(0);
    }

    if (stat(filename, &stat_buf) == -1) {
      log("unable to stat %s: %s\n", filename, strerror(errno));
      assert(0);
    }

    num_events = (stat_buf.st_size / sizeof(struct event)) / 2;
  }

  /* Allocate space for the ev_arrive and ev_complete arrays, then
   * proceed through the file filling in the arrays. */

  ev_arrive = (struct event *)calloc(num_events, sizeof(struct event));
  ev_complete = (struct event *)calloc(num_events, sizeof(struct event));
  assert(ev_arrive && ev_complete);

  while (1) {
    int rc;
    struct event ev;

    rc = read(fd, &ev, sizeof(struct event));
    if (rc != sizeof(struct event)) {
      if (rc == -1 && errno == EAGAIN) continue;
      if (rc > 0) log("short read: %d < %d\n", rc, sizeof(struct event));
      break;
    }

    if (ev.opcode != 0x28 && ev.opcode != 0x2a
	&& ev.opcode != 0x08 && ev.opcode != 0x0a) continue;

    switch (ev.code) {

    case MEMULATOR_EVENT_HOST_KERNEL_ARRIVE:
      assert(ev_arrive_index < num_events);
      memcpy(&ev_arrive[ev_arrive_index++], &ev, sizeof(struct event));
      break;

    case MEMULATOR_EVENT_HOST_KERNEL_COMPLETE:
      assert(ev_complete_index < num_events);
      memcpy(&ev_complete[ev_complete_index++], &ev, sizeof(struct event));
      break;

    default:
      log("unknown code 0x%x (%d:%d, opcode 0x%d, ts %lld\n", ev.code, ev.offset, ev.length, ev.opcode, ev.u.tsc);
    }
  }

  close(fd);
  assert(ev_arrive_index == ev_complete_index);

  /* Match each of the entries in the ev_arrive array with the entries
   * in the ev_complete array (it is possible for the entries to be out
   * of order between the arrays), and create the timeinfo array that
   * will be exported outside of this function. */
  {
    int a_index = 0;
    int c_index = 0;  /* Also used as the index into the timeinfo array */

    ti_hos = (struct timeinfo *)calloc(num_events, sizeof(struct timeinfo));
    assert(ti_hos);

    for (c_index = 0; c_index < ev_complete_index; c_index++) {
      int i;

      ti_hos[c_index].offset = ev_complete[c_index].offset;
      ti_hos[c_index].length = ev_complete[c_index].length;
      ti_hos[c_index].opcode = ev_complete[c_index].opcode;
      ti_hos[c_index].hos_complete = ev_complete[c_index].u.tsc;

      while (ev_arrive[a_index].code == 0) a_index++;

      for (i = a_index; i < ev_arrive_index; i++) {
	if (ev_arrive[i].code == 0) continue;
	if (ev_arrive[i].offset == ti_hos[c_index].offset
	    && ev_arrive[i].length == ti_hos[c_index].length
	    && ev_arrive[i].opcode == ti_hos[c_index].opcode) {
	  ti_hos[c_index].hos_arrive = ev_arrive[i].u.tsc;
	  ev_arrive[i].code = 0;
	  break;
	}
      }
      assert(i < ev_arrive_index);  /* we must find a match, or die trying */

      ti_hos_num_entries++;
    }
    assert(c_index == ti_hos_num_entries);
  }

  free(ev_arrive); ev_arrive = NULL;
  free(ev_complete); ev_complete = NULL;
}


void
process_mem (void)
{
  int fd;
  int num_events;
  struct event *ev_arrive, *ev_complete;
  int ev_arrive_index = 0, ev_complete_index = 0;

  /* Open the file, use the stat information to determine the length
   * of the file, then use that to determine the length of the
   * ev_arrive and ev_complete arrays that we need to allocate. */
  {
    const int filename_maxlen = 80;
    char filename[filename_maxlen];
    struct stat stat_buf;

    assert((strlen(prefix) + 4) < filename_maxlen);
    snprintf(filename, filename_maxlen, "%s.mem", prefix);

    fd = open(filename, O_RDONLY);
    if (fd == -1) {
      log("unable to open %s: %s\n", filename, strerror(errno));
      assert(0);
    }

    if (stat(filename, &stat_buf) == -1) {
      log("unable to stat %s: %s\n", filename, strerror(errno));
      assert(0);
    }

    num_events = (stat_buf.st_size / sizeof(struct event)) / 2;
  }

  /* Allocate space for the ev_arrive and ev_complete arrays, then
   * proceed through the file filling in the arrays. */

  ev_arrive = (struct event *)calloc(num_events, sizeof(struct event));
  ev_complete = (struct event *)calloc(num_events, sizeof(struct event));
  assert(ev_arrive && ev_complete);

  while (1) {
    int rc;
    struct event ev;

    rc = read(fd, &ev, sizeof(struct event));
    if (rc != sizeof(struct event)) {
      if (rc == -1 && errno == EAGAIN) continue;
      if (rc > 0) log("short read: %d < %d\n", rc, sizeof(struct event));
      break;
    }

    if (ev.opcode != 0x28 && ev.opcode != 0x2a
	&& ev.opcode != 0x08 && ev.opcode != 0x0a) continue;

    switch (ev.code) {

    case MEMULATOR_EVENT_TARGET_EMULATOR_ARRIVE:
      assert(ev_arrive_index < num_events);
      memcpy(&ev_arrive[ev_arrive_index++], &ev, sizeof(struct event));
      break;

    case MEMULATOR_EVENT_TARGET_EMULATOR_COMPLETE:
      assert(ev_complete_index < num_events);
      memcpy(&ev_complete[ev_complete_index++], &ev, sizeof(struct event));
      break;

    default:
      log("unknown code 0x%x (%d:%d, opcode 0x%d, ts %lld\n", ev.code, ev.offset, ev.length, ev.opcode, ev.u.tsc);
    }
  }

  close(fd);
  assert(ev_arrive_index == ev_complete_index);

  /* Match each of the entries in the ev_arrive array with the entries
   * in the ev_complete array (it is possible for the entries to be out
   * of order between the arrays), and create the timeinfo array that
   * will be exported outside of this function. */
  {
    int a_index = 0;
    int c_index = 0;  /* Also used as the index into the timeinfo array */

    ti_mem = (struct timeinfo *)calloc(num_events, sizeof(struct timeinfo));
    assert(ti_mem);

    for (c_index = 0; c_index < ev_complete_index; c_index++) {
      int i;

      ti_mem[c_index].offset = ev_complete[c_index].offset;
      ti_mem[c_index].length = ev_complete[c_index].length;
      ti_mem[c_index].opcode = ev_complete[c_index].opcode;
      ti_mem[c_index].mem_complete = ev_complete[c_index].u.tsc;

      while (ev_arrive[a_index].code == 0) a_index++;

      for (i = a_index; i < ev_arrive_index; i++) {
	if (ev_arrive[i].code == 0) continue;
	if (ev_arrive[i].offset == ti_mem[c_index].offset
	    && ev_arrive[i].length == ti_mem[c_index].length
	    && ev_arrive[i].opcode == ti_mem[c_index].opcode) {
	  ti_mem[c_index].mem_arrive = ev_arrive[i].u.tsc;
	  ev_arrive[i].code = 0;
	  break;
	}
      }
      assert(i < ev_arrive_index);  /* we must find a match, or die trying */

      ti_mem_num_entries++;
    }
    assert(c_index == ti_mem_num_entries);
  }

  free(ev_arrive); ev_arrive = NULL;
  free(ev_complete); ev_complete = NULL;
}


void
process_sim (void)
{
  int fd;
  int num_events;

  /* Open the file, use the stat information to determine the length
   * of the file, then use that to determine the length of the
   * ev_arrive and ev_complete arrays that we need to allocate. */
  {
    const int filename_maxlen = 80;
    char filename[filename_maxlen];
    struct stat stat_buf;

    assert((strlen(prefix) + 4) < filename_maxlen);
    snprintf(filename, filename_maxlen, "%s.sim", prefix);

    fd = open(filename, O_RDONLY);
    if (fd == -1) {
      log("unable to open %s: %s\n", filename, strerror(errno));
      assert(0);
    }

    if (stat(filename, &stat_buf) == -1) {
      log("unable to stat %s: %s\n", filename, strerror(errno));
      assert(0);
    }

    num_events = (stat_buf.st_size / sizeof(struct event));
  }

  /* Allocate space for the ev_arrive and ev_complete arrays, then
   * proceed through the file filling in the arrays. */

  ti_sim = (struct timeinfo *)calloc(num_events, sizeof(struct timeinfo));
  assert(ti_sim);

  while (1) {
    int rc;
    struct event ev;

    rc = read(fd, &ev, sizeof(struct event));
    if (rc != sizeof(struct event)) {
      if (rc == -1 && errno == EAGAIN) continue;
      if (rc > 0) log("short read: %d < %d\n", rc, sizeof(struct event));
      break;
    }

    switch (ev.code) {

    case MEMULATOR_EVENT_TARGET_SIMULATOR_TIME:
      assert(ti_sim_num_entries < num_events);
      ti_sim[ti_sim_num_entries].offset = ev.offset;
      ti_sim[ti_sim_num_entries].length = ev.length;
      ti_sim[ti_sim_num_entries].opcode = ev.opcode;
      ti_sim[ti_sim_num_entries].sim_time = ev.u.elapsed / 1000.0;
      ti_sim_num_entries++;
      break;

    default:
      log("unknown code 0x%x (%d:%d, opcode 0x%d, ts %f\n", ev.code, ev.offset, ev.length, ev.opcode, ev.u.elapsed);
    }
  }

  close(fd);
}


void
merge_hos_mem (void)
{
  int hos_index;
  int mem_index;

  for (hos_index = 0, mem_index = 0;
       hos_index < ti_hos_num_entries;
       hos_index++, mem_index++) {

    while (ti_hos[hos_index].offset != ti_mem[mem_index].offset
	   || ti_hos[hos_index].length != ti_mem[mem_index].length
	   || ti_hos[hos_index].opcode != ti_mem[mem_index].opcode) {
      log("skipping mem entry %d\n", mem_index);
      mem_index++;
      assert(mem_index < ti_mem_num_entries);
    }

    ti_hos[hos_index].mem_arrive   = ti_mem[mem_index].mem_arrive;
    ti_hos[hos_index].mem_complete = ti_mem[mem_index].mem_complete;
  }
}


void
merge_hos_sim (void)
{
  int hos_index;
  int sim_index;

  for (hos_index = 0, sim_index = 0;
       hos_index < ti_hos_num_entries;
       hos_index++, sim_index++) {

    while (ti_hos[hos_index].offset != ti_sim[sim_index].offset
	   || ti_hos[hos_index].length != ti_sim[sim_index].length
	   /* GRONK: fix opcode? */) {
      log("skipping sim entry %d\n", sim_index);
      sim_index++;
      assert(sim_index < ti_sim_num_entries);
    }

    ti_hos[hos_index].sim_time = ti_sim[sim_index].sim_time;
  }
}


void
merge_mem_sim (void)
{
  int mem_index;
  int sim_index;

  for (mem_index = 0, sim_index = 0;
       mem_index < ti_mem_num_entries;
       mem_index++, sim_index++) {

    while (ti_mem[mem_index].offset != ti_sim[sim_index].offset
	   || ti_mem[mem_index].length != ti_sim[sim_index].length
	   /* GRONK: fix opcode? */) {
      log("skipping sim entry %d\n", sim_index);
      sim_index++;
      assert(sim_index < ti_sim_num_entries);
    }

    ti_mem[mem_index].sim_time = ti_sim[sim_index].sim_time;
  }
}


void
hos_mem_analyze (void)
{
  int i;
  double min_diff=999999.99;
  double max_diff=-999999.99;
  double total_diff=0.0;

  double nc93 = 1993907171.0;
  double nc94 = 1993868287.0;

  for (i = 0; i < ti_hos_num_entries; i++) {
    u_int64_t hos_delta, mem_delta;
    double hos_time, mem_time;
    double diff;
    u_int64_t diff2;

    hos_delta = ti_hos[i].hos_complete - ti_hos[i].hos_arrive;
    mem_delta = ti_hos[i].mem_complete - ti_hos[i].mem_arrive;

    hos_time = hos_delta / nc93;
    mem_time = mem_delta / nc94;

    diff = hos_time - mem_time;
    diff2 = hos_delta - mem_delta;

    //printf("%3d: (%8d:%3d) diff %f diff2 %lld \n", i, ti_hos[i].offset, ti_hos[i].length, diff, diff2);

    total_diff += diff;
    if (diff < min_diff) min_diff = diff;
    if (diff > max_diff) max_diff = diff;
  }

  hos_mem_avg = total_diff / (double)i;
  hos_mem_min = min_diff;
  hos_mem_max = max_diff;
  hos_mem_stddev = -1.0;
}


void
hos_sim_analyze (void)
{
  int i;
  double min_diff=999999.99;
  double max_diff=-999999.99;
  double total_diff=0.0;

  double nc93 = 1993907171.0;

  for (i = 0; i < ti_hos_num_entries; i++) {
    u_int64_t hos_delta;
    double hos_time;
    double diff;

    hos_delta = ti_hos[i].hos_complete - ti_hos[i].hos_arrive;

    hos_time = hos_delta / nc93;

    diff = hos_time - ti_hos[i].sim_time;

    //printf("%3d: (%8d:%3d) hos %f, sim %f, diff %f\n", i, ti_hos[i].offset, ti_hos[i].length, hos_time, ti_hos[i].sim_time, diff);

    total_diff += diff;
    if (diff < min_diff) min_diff = diff;
    if (diff > max_diff) max_diff = diff;
  }

  hos_sim_avg = total_diff / (double)i;
  hos_sim_min = min_diff;
  hos_sim_max = max_diff;
  hos_sim_stddev = -1.0;
}


void
mem_sim_analyze (void)
{
  int i;
  double min_diff=999999.99;
  double max_diff=-999999.99;
  double total_diff=0.0;

  double nc93 = 1993907171.0;

  for (i = 0; i < ti_mem_num_entries; i++) {
    u_int64_t mem_delta;
    double mem_time;
    double diff;

    mem_delta = ti_mem[i].mem_complete - ti_mem[i].mem_arrive;

    mem_time = mem_delta / nc93;

    diff = mem_time - ti_mem[i].sim_time;

    //printf("%3d: (%8d:%3d) mem %f, sim %f, diff %f\n", i, ti_mem[i].offset, ti_mem[i].length, mem_time, ti_mem[i].sim_time, diff);

    total_diff += diff;
    if (diff < min_diff) min_diff = diff;
    if (diff > max_diff) max_diff = diff;
  }

  mem_sim_avg = total_diff / (double)i;
  mem_sim_min = min_diff;
  mem_sim_max = max_diff;
  mem_sim_stddev = -1.0;
}


void
process_args (int argc, char **argv)
{
  char *program = argv[0];
  int ch;

  while ((ch = getopt(argc, argv, "f:s:")) != -1) {
    switch (ch) {
    case 'f':  /* file prefix */
      prefix = optarg;
      break;
    case 's':
      output_string = optarg;
      break;
    default:
      goto usage;
    }
  }
  argc -= optind;
  argv += optind;

  if (argc != 0) {
    printf("argc==%d\n", argc);
  usage:
    printf("usage:\n"
	   "\t%s -f prefix\n"
	   , program);
    exit(1);
  }
}


int
main (int argc, char **argv)
{
  process_args(argc, argv);

  process_hos();
  process_mem();
  process_sim();

  if (1) {
    merge_hos_mem();
    hos_mem_analyze();
  }

  if (1) {
    merge_hos_sim();
    hos_sim_analyze();
  }

  if (1) {
    merge_mem_sim();
    mem_sim_analyze();
  }

  printf("%s %f %f %f %f %f\n",
	 output_string,
	 hos_sim_avg, hos_sim_stddev,
	 mem_sim_avg, mem_sim_stddev,
	 -1.0);

  printf("done\n");

  return 0;
}
